#!/usr/bin/env python
"""This module has unit tests for the pvl __init__ functions."""

# Copyright 2019, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import call, create_autospec, mock_open, patch

from pathlib import Path

import pvl

data_dir = Path("tests/data")


class TestLoadS(unittest.TestCase):
    def test_loads(self):
        some_pvl = """
a = b
GROUP = c
    c = d
END_GROUP
e =false
END"""
        decoded = pvl.PVLModule(a="b", c=pvl.PVLGroup(c="d"), e=False)
        self.assertEqual(decoded, pvl.loads(some_pvl))

        self.assertEqual(pvl.PVLModule(a="b"), pvl.loads("a=b"))


class TestLoad(unittest.TestCase):
    def setUp(self):
        self.simple = data_dir / "pds3" / "simple_image_1.lbl"
        rawurl = "https://raw.githubusercontent.com/planetarypy/pvl/main/"
        self.url = rawurl + str(self.simple)
        self.simplePVL = pvl.PVLModule(
            {
                "PDS_VERSION_ID": "PDS3",
                "RECORD_TYPE": "FIXED_LENGTH",
                "RECORD_BYTES": 824,
                "LABEL_RECORDS": 1,
                "FILE_RECORDS": 601,
                "^IMAGE": 2,
                "IMAGE": pvl.PVLObject(
                    {
                        "LINES": 600,
                        "LINE_SAMPLES": 824,
                        "SAMPLE_TYPE": "MSB_INTEGER",
                        "SAMPLE_BITS": 8,
                        "MEAN": 51.67785396440129,
                        "MEDIAN": 50.0,
                        "MINIMUM": 0,
                        "MAXIMUM": 255,
                        "STANDARD_DEVIATION": 16.97019,
                        "CHECKSUM": 25549531,
                    }
                ),
            }
        )

    def test_load_w_open(self):
        with open(self.simple) as f:
            self.assertEqual(self.simplePVL, pvl.load(f))

    def test_load_w_Path(self):
        self.assertEqual(self.simplePVL, pvl.load(self.simple))

    def test_load_w_string_path(self):
        string_path = str(self.simple)
        self.assertEqual(self.simplePVL, pvl.load(string_path))

    def test_loadu(self):
        self.assertEqual(self.simplePVL, pvl.loadu(self.url))
        self.assertEqual(
            self.simplePVL, pvl.loadu(self.simple.resolve().as_uri())
        )

    @patch("pvl.loads")
    @patch("pvl.decode_by_char")
    def test_loadu_args(self, m_decode, m_loads):
        pvl.loadu(self.url, data=None)
        pvl.loadu(self.url, noturlopen="should be passed to loads")
        m_decode.assert_called()
        self.assertNotIn("data", m_loads.call_args_list[0][1])
        self.assertIn("noturlopen", m_loads.call_args_list[1][1])

    def test_load_w_quantity(self):
        try:
            from astropy import units as u
            from pvl.decoder import OmniDecoder

            pvl_file = "tests/data/pds3/units1.lbl"
            km_upper = u.def_unit("KM", u.km)
            m_upper = u.def_unit("M", u.m)
            u.add_enabled_units([km_upper, m_upper])
            label = pvl.load(
                pvl_file, decoder=OmniDecoder(quantity_cls=u.Quantity)
            )
            self.assertEqual(label["FLOAT_UNIT"], u.Quantity(0.414, "KM"))
        except ImportError:
            pass


class TestISIScub(unittest.TestCase):
    def setUp(self):
        self.cub = data_dir / "pattern.cub"
        self.cubpvl = pvl.PVLModule(
            IsisCube=pvl.PVLObject(
                Core=pvl.PVLObject(
                    StartByte=65537,
                    Format="Tile",
                    TileSamples=128,
                    TileLines=128,
                    Dimensions=pvl.PVLGroup(Samples=90, Lines=90, Bands=1),
                    Pixels=pvl.PVLGroup(
                        Type="Real", ByteOrder="Lsb", Base=0.0, Multiplier=1.0
                    ),
                )
            ),
            Label=pvl.PVLObject(Bytes=65536),
        )

    def test_load_cub(self):
        self.assertEqual(self.cubpvl, pvl.load(self.cub))

    def test_load_cub_opened(self):
        with open(self.cub, "rb") as f:
            self.assertEqual(self.cubpvl, pvl.load(f))


class TestDumpS(unittest.TestCase):
    def setUp(self):
        self.module = pvl.PVLModule(
            a="b",
            staygroup=pvl.PVLGroup(c="d"),
            obj=pvl.PVLGroup(d="e", f=pvl.PVLGroup(g="h")),
        )

    def test_dumps_PDS(self):
        s = """A = b\r
GROUP = staygroup\r
  C = d\r
END_GROUP = staygroup\r
OBJECT = obj\r
  D = e\r
  GROUP = f\r
    G = h\r
  END_GROUP = f\r
END_OBJECT = obj\r
END\r\n"""
        self.assertEqual(s, pvl.dumps(self.module))

    def test_dumps_PVL(self):
        s = """a = b;
BEGIN_GROUP = staygroup;
  c = d;
END_GROUP = staygroup;
BEGIN_GROUP = obj;
  d = e;
  BEGIN_GROUP = f;
    g = h;
  END_GROUP = f;
END_GROUP = obj;
END;"""

        self.assertEqual(
            s, pvl.dumps(self.module, encoder=pvl.encoder.PVLEncoder())
        )

    def test_dumps_ODL(self):

        s = """A = b\r
GROUP = staygroup\r
  C = d\r
END_GROUP = staygroup\r
GROUP = obj\r
  D = e\r
  GROUP = f\r
    G = h\r
  END_GROUP = f\r
END_GROUP = obj\r
END\r\n"""

        self.assertEqual(
            s, pvl.dumps(self.module, encoder=pvl.encoder.ODLEncoder())
        )


class TestDump(unittest.TestCase):
    def setUp(self):
        self.module = pvl.PVLModule(
            a="b",
            staygroup=pvl.PVLGroup(c="d"),
            obj=pvl.PVLGroup(d="e", f=pvl.PVLGroup(g="h")),
        )
        self.string = """A = b\r
GROUP = staygroup\r
  C = d\r
END_GROUP = staygroup\r
OBJECT = obj\r
  D = e\r
  GROUP = f\r
    G = h\r
  END_GROUP = f\r
END_OBJECT = obj\r
END\r\n"""

    def test_dump_Path(self):
        mock_path = create_autospec(Path)
        with patch("pvl.Path", autospec=True, return_value=mock_path):
            pvl.dump(self.module, Path("dummy"))
            self.assertEqual(
                [call.write_text(self.string)], mock_path.method_calls
            )

    @patch("builtins.open", mock_open())
    def test_dump_file_object(self):
        with open("dummy", "w") as f:
            pvl.dump(self.module, f)
            self.assertEqual(
                [call.write(self.string.encode())], f.method_calls
            )

    def test_not_dumpable(self):
        f = 5
        self.assertRaises(TypeError, pvl.dump, self.module, f)



# -*- coding: utf-8 -*-
"""A program for testing and validating PVL text.

The ``pvl_validate`` program will read a file with PVL text (any of
the kinds of files that :func:`pvl.load` reads) and will report
on which of the various PVL dialects were able to load that PVL
text, and then also reports on whether the ``pvl`` library can encode
the Python Objects back out to PVL text.

You can imagine some PVL text that could be loaded, but is not able
to be written out in a particular strict PVL dialect (like PDS3
labels).
"""

# Copyright 2020-2021, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
import sys
from collections import OrderedDict

import pvl
from .lexer import LexerError
from .grammar import (
    PVLGrammar,
    ODLGrammar,
    PDSGrammar,
    ISISGrammar,
    OmniGrammar,
)
from .parser import ParseError, PVLParser, ODLParser, OmniParser
from .decoder import PVLDecoder, ODLDecoder, PDSLabelDecoder, OmniDecoder
from .encoder import PVLEncoder, ODLEncoder, ISISEncoder, PDSLabelEncoder

# Some assembly required for the dialects.
# We are going to be explicit here, because these arguments are
# are different than the defaults for these classes, especially for the
# parsers and decoders, as we want to be strict and not permissive here.
_pvl_g = PVLGrammar()
_pvl_d = PVLDecoder(grammar=_pvl_g)
_odl_g = ODLGrammar()
_odl_d = ODLDecoder(grammar=_odl_g)
_pds_g = PDSGrammar()
_pds_d = PDSLabelDecoder(grammar=_pds_g)
_isis_g = ISISGrammar()
_isis_d = OmniDecoder(grammar=_isis_g)
_omni_g = OmniGrammar()
_omni_d = OmniDecoder(grammar=_omni_g)

dialects = OrderedDict(
    PDS3=dict(
        parser=ODLParser(grammar=_pds_g, decoder=_pds_d),
        grammar=_pds_g,
        decoder=_pds_d,
        encoder=PDSLabelEncoder(grammar=_pds_g, decoder=_pds_d),
    ),
    ODL=dict(
        parser=ODLParser(grammar=_odl_g, decoder=_odl_d),
        grammar=_odl_g,
        decoder=_odl_d,
        encoder=ODLEncoder(grammar=_odl_g, decoder=_odl_d),
    ),
    PVL=dict(
        parser=PVLParser(grammar=_pvl_g, decoder=_pvl_d),
        grammar=_pvl_g,
        decoder=_pvl_d,
        encoder=PVLEncoder(grammar=_pvl_g, decoder=_pvl_d),
    ),
    ISIS=dict(
        parser=OmniParser(grammar=_isis_g, decoder=_isis_d),
        grammar=_isis_g,
        decoder=_isis_d,
        encoder=ISISEncoder(grammar=_isis_g, decoder=_isis_d),
    ),
    Omni=dict(
        parser=OmniParser(grammar=_omni_g, decoder=_omni_d),
        grammar=_omni_g,
        decoder=_omni_d,
        encoder=PVLEncoder(grammar=_omni_g, decoder=_omni_d),
    ),
)


def arg_parser():
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Will report the errors that are encountered.  A second v will "
        "include tracebacks for non-pvl exceptions. ",
    )
    p.add_argument("--version", action="version", version=pvl.__version__)
    p.add_argument(
        "file", nargs="+", help="file containing PVL text to validate."
    )
    return p


def main():
    args = arg_parser().parse_args()

    logging.basicConfig(
        format="%(levelname)s: %(message)s", level=(60 - 20 * args.verbose)
    )

    results_list = list()
    for f in args.file:
        pvl_text = pvl.get_text_from(f)

        results = dict()

        for k, v in dialects.items():
            results[k] = pvl_flavor(pvl_text, k, v, f, args.verbose)

        results_list.append((f, results))

    # Writing the flavors out again to preserve order.
    if args.verbose > 0:
        print(f"pvl library version: {pvl.__version__}")
    print(report(results_list, list(dialects.keys())))
    return


def pvl_flavor(
    text, dialect, decenc: dict, filename, verbose=False
) -> tuple((bool, bool)):
    """Returns a two-tuple of booleans which indicate
    whether the *text* could be loaded and then encoded.

    The first boolean in the two-tuple indicates whether the *text*
    could be loaded with the given parser, grammar, and decoder.
    The second indicates whether the loaded PVL object could be
    encoded with the given encoder, grammar, and decoder.  If the
    first element is False, the second will be None.
    """
    loads = None
    encodes = None
    try:
        some_pvl = pvl.loads(text, **decenc)
        loads = True

        try:
            pvl.dumps(some_pvl, **decenc)
            encodes = True
        except (LexerError, ParseError, ValueError) as err:
            logging.error(f"{dialect} encode error {filename} {err}")
            encodes = False
    except (LexerError, ParseError) as err:
        logging.error(f"{dialect} load error {filename} {err}")
        loads = False
    except:  # noqa E722
        if verbose <= 1:
            logging.error(
                f"{dialect} load error {filename}, try -vv for more info."
            )
        else:
            logging.exception(f"{dialect} load error {filename}")
            logging.error(f"End {dialect} load error {filename}")
        loads = False

    return loads, encodes


def report(reports: list, flavors: list) -> str:
    """Returns a multi-line string which is the
    pretty-printed report given the list of
    *reports*.
    """
    if len(reports[0][1]) != len(flavors):
        raise IndexError(
            "The length of the report list keys "
            f"({len(reports[0][1])}) "
            "and the length of the flavors list "
            f"({len(flavors)}) aren't the same."
        )

    if len(reports) > 1:
        return report_many(reports, flavors)

    r = reports[0][1]

    lines = list()
    loads = {True: "Loads", False: "does NOT load"}
    encodes = {True: "Encodes", False: "does NOT encode", None: ""}

    col1w = len(max(flavors, key=len))
    col2w = len(max(loads.values(), key=len))
    col3w = len(max(encodes.values(), key=len))

    for k in flavors:
        lines.append(
            build_line(
                [k, loads[r[k][0]], encodes[r[k][1]]], [col1w, col2w, col3w]
            )
        )
    return "\n".join(lines)


def report_many(r_list: list, flavors: list) -> str:
    """Returns a multi-line, table-like string which
    is the pretty-printed report of the items in *r_list*.
    """

    lines = list()
    loads = {True: "L", False: "No L"}
    encodes = {True: "E", False: "No E", None: ""}

    col1w = len(max([x[0] for x in r_list], key=len))
    col2w = len(max(loads.values(), key=len))
    col3w = len(max(encodes.values(), key=len))
    flavorw = col2w + col3w + 1

    header = ["File"] + flavors
    headerw = [col1w] + [flavorw] * len(flavors)
    rule = [" " * col1w] + [" " * flavorw] * len(flavors)

    rule_line = build_line(rule, headerw).replace("|", "+").replace(" ", "-")
    lines.append(rule_line)
    lines.append(build_line(header, headerw))
    lines.append(rule_line)

    for r in r_list:
        cells = [r[0]]
        widths = [col1w]
        for f in flavors:
            # cells.append(loads[r[1][f][0]] + ' ' + encodes[r[1][f][1]])
            cells.append(
                "{0:^{w2}} {1:^{w3}}".format(
                    loads[r[1][f][0]], encodes[r[1][f][1]], w2=col2w, w3=col3w
                )
            )
            widths.append(flavorw)
        lines.append(build_line(cells, widths))

    return "\n".join(lines)


def build_line(elements: list, widths: list, sep=" | ") -> str:
    """Returns a string formatted from the *elements* and *widths*
       provided.
    """
    cells = list()
    cells.append("{0:<{width}}".format(elements[0], width=widths[0]))

    for e, w in zip(elements[1:], widths[1:]):
        cells.append("{0:^{width}}".format(e, width=w))

    return sep.join(cells)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Provides lexer functions for PVL."""

# Copyright 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.


from enum import Enum, auto

from .grammar import PVLGrammar
from .token import Token
from .decoder import PVLDecoder
from .exceptions import LexerError, firstpos


class Preserve(Enum):
    FALSE = auto()
    COMMENT = auto()
    UNIT = auto()
    QUOTE = auto()
    NONDECIMAL = auto()


def lex_preserve(char: str, lexeme: str, preserve: dict) -> tuple((str, dict)):
    """Returns a modified *lexeme* string and a modified *preserve*
    dict in a two-tuple.  The modified *lexeme* will always be
    the concatenation of *lexeme* and *char*.

    This is a lexer() helper function that is responsible for
    changing the state of the *preserve* dict, if needed.

    If the value for 'end' in *preserve* is the same as *char*,
    then the modified *preserve* will have its 'state' value
    set to ``Preserve.FALSE`` and its 'end' value set to None,
    otherwise second item in the returned tuple will be *preserve*
    unchanged.
    """
    # print(f'in preserve: char "{char}", lexeme "{lexeme}, p {preserve}"')
    if char == preserve["end"]:
        return lexeme + char, dict(state=Preserve.FALSE, end=None)
    else:
        return lexeme + char, preserve


def lex_singlechar_comments(
    char: str, lexeme: str, preserve: dict, comments: dict
) -> tuple((str, dict)):
    """Returns a modified *lexeme* string and a modified *preserve*
    dict in a two-tuple.

    This is a lexer() helper function for determining how to modify
    *lexeme* and *preserve* based on the single character in *char*
    which may or may not be a comment character.

    If the *preserve* 'state' value is Preserve.COMMENT then
    the value of lex_preserve() is returned.

    If *char* is among the keys of the *comments* dict, then the
    returned *lexeme* will be the concatenation of *lexeme* and
    *char*.  returned *preserve* dict will have its 'state' value
    set to Preserve.COMMENT and its 'end' value set to the value
    of *comments[char]*.

    Otherwise return *lexeme* and *preserve* unchanged in the
    two-tuple.
    """
    if preserve["state"] == Preserve.COMMENT:
        return lex_preserve(char, lexeme, preserve)
    elif char in comments:
        return (
            lexeme + char,
            dict(state=Preserve.COMMENT, end=comments[char]),
        )

    return lexeme, preserve


def lex_multichar_comments(
    char: str,
    prev_char: str,
    next_char: str,
    lexeme: str,
    preserve: dict,
    comments: tuple(tuple((str, str))) = PVLGrammar().comments,
) -> tuple((str, dict)):
    """Returns a modified *lexeme* string and a modified *preserve*
    dict in a two-tuple.

    This is a lexer() helper function for determining how to
    modify *lexeme* and *preserve* based on the single character
    in *char* which may or may not be part of a multi-character
    comment character group.

    This function has an internal list of allowed pairs of
    multi-character comments that it can deal with, if the
    *comments* tuple contains any two-tuples that cannot be
    handled, a NotImplementedError will be raised.

    This function will determine whether to append *char* to
    *lexeme* or not, and will set the value of the 'state' and
    'end' values of *preserve* appropriately.
    """
    # print(f'lex_multichar got these comments: {comments}')
    if len(comments) == 0:
        raise ValueError("The variable provided to comments is empty.")

    allowed_pairs = (("/*", "*/"),)
    for p in comments:
        if p not in allowed_pairs:
            raise NotImplementedError(
                "Can only handle these "
                "multicharacter comments: "
                f"{allowed_pairs}.  To handle "
                "others this class must be extended."
            )

    if ("/*", "*/") in comments:
        if char == "*":
            if prev_char == "/":
                return lexeme + "/*", dict(state=Preserve.COMMENT, end="*/")
            elif next_char == "/":
                return lexeme + "*/", dict(state=Preserve.FALSE, end=None)
            else:
                return lexeme + "*", preserve
        elif char == "/":
            # If part of a comment ignore, and let the char == '*' handler
            # above deal with it, otherwise add it to the lexeme.
            if prev_char != "*" and next_char != "*":
                return lexeme + "/", preserve

    return lexeme, preserve


def lex_comment(
    char: str,
    prev_char: str,
    next_char: str,
    lexeme: str,
    preserve: dict,
    c_info: dict,
) -> tuple((str, dict)):
    """Returns a modified *lexeme* string and a modified *preserve*
    dict in a two-tuple.

    This is a lexer() helper function for determining how to
    modify *lexeme* and *preserve* based on the single character
    in *char* which may or may not be a comment character.

    This function just makes the decision about whether to call
    lex_multichar_comments() or lex_singlechar_comments(), and
    then returns what they return.
    """

    if char in c_info["multi_chars"]:
        return lex_multichar_comments(
            char,
            prev_char,
            next_char,
            lexeme,
            preserve,
            comments=c_info["multi_comments"],
        )
    else:
        return lex_singlechar_comments(
            char, lexeme, preserve, c_info["single_comments"]
        )


def _prev_char(s: str, idx: int):
    """Returns the character from *s* at the position before *idx*
    or None, if *idx* is zero.
    """
    if idx <= 0:
        return None
    else:
        return s[idx - 1]


def _next_char(s: str, idx: int):
    """Returns the character from *s* at the position after *idx*
    or None, if *idx* is the last position in *s*.
    """
    try:
        return s[idx + 1]
    except IndexError:
        return None


def _prepare_comment_tuples(comments: tuple(tuple((str, str)))) -> dict:
    """Returns a dict of information based on the contents
    of *comments*.

    This is a lexer() helper function to prepare information
    for lexer().
    """
    # I initially tried to avoid this function, if you
    # don't pre-compute this stuff, you end up re-computing
    # it every time you pass into the lex_comment() function,
    # which seemed excessive.
    d = dict()
    m = list()
    d["single_comments"] = dict()
    d["multi_chars"] = set()
    for pair in comments:
        if len(pair[0]) == 1:
            d["single_comments"][pair[0]] = pair[1]
        else:
            m.append(pair)
            for p in pair:
                d["multi_chars"] |= set(p)

    d["chars"] = set(d["single_comments"].keys())
    d["chars"] |= d["multi_chars"]
    d["multi_comments"] = tuple(m)

    # print(d)
    return d


def lex_char(
    char: str,
    prev_char: str,
    next_char: str,
    lexeme: str,
    preserve: dict,
    g: PVLGrammar,
    c_info: dict,
) -> tuple((str, dict)):
    """Returns a modified *lexeme* string and a modified *preserve*
    dict in a two-tuple.

    This is the main lexer() helper function for determining how
    to modify (or not) *lexeme* and *preserve* based on the
    single character in *char* and the other values passed into
    this function.
    """

    # When we are 'in' a comment or a units expression,
    # we want those to consume everything, regardless.
    # So we must handle the 'preserve' states first,
    # and then after that we can check to see if the char
    # should put us into one of those states.

    # print(f'lex_char start: char "{char}", lexeme "{lexeme}", "{preserve}"')

    if preserve["state"] != Preserve.FALSE:
        if preserve["state"] == Preserve.COMMENT:
            (lexeme, preserve) = lex_comment(
                char, prev_char, next_char, lexeme, preserve, c_info
            )
        elif preserve["state"] in (
            Preserve.UNIT,
            Preserve.QUOTE,
            Preserve.NONDECIMAL,
        ):
            (lexeme, preserve) = lex_preserve(char, lexeme, preserve)
        else:
            raise ValueError(
                "{} is not a ".format(preserve["state"])
                + "recognized preservation state."
            )
    elif (
        char == "#"
        and g.nondecimal_pre_re.fullmatch(lexeme + char) is not None
    ):
        lexeme += char
        preserve = dict(state=Preserve.NONDECIMAL, end="#")
    elif char in c_info["chars"]:
        (lexeme, preserve) = lex_comment(
            char, prev_char, next_char, lexeme, preserve, c_info
        )
    elif char in g.units_delimiters[0]:
        lexeme += char
        preserve = dict(state=Preserve.UNIT, end=g.units_delimiters[1])
    elif char in g.quotes:
        lexeme += char
        preserve = dict(state=Preserve.QUOTE, end=char)
    else:
        if char not in g.whitespace:
            lexeme += char  # adding a char each time

    # print(f'lex_char end: char "{char}", lexeme "{lexeme}", "{preserve}"')
    return lexeme, preserve


def lex_continue(
    char: str,
    next_char: str,
    lexeme: str,
    token: Token,
    preserve: dict,
    g: PVLGrammar,
) -> bool:
    """Return True if accumulation of *lexeme* should continue based
    on the values passed into this function, false otherwise.

    This is a lexer() helper function.
    """

    if next_char is None:
        return False

    if not g.char_allowed(next_char):
        return False

    if preserve["state"] != Preserve.FALSE:
        return True

    # Since Numeric objects can begin with a reserved
    # character, the reserved characters may split up
    # the lexeme.
    if (
        char in g.numeric_start_chars
        and Token(char + next_char, grammar=g).is_numeric()
    ):
        return True

    # Since Non Decimal Numerics can have reserved characters in them.
    if g.nondecimal_pre_re.fullmatch(lexeme + next_char) is not None:
        return True

    # Since the numeric signs could be in the reserved characters,
    # make sure we can parse scientific notation correctly:
    if (
        char.lower() == "e"
        and next_char in g.numeric_start_chars
        and Token(lexeme + next_char + "2", grammar=g).is_numeric()
    ):
        return True

    # Some datetimes can have trailing numeric tz offsets,
    # if the decoder allows it, this means there could be
    # a '+' that splits the lexeme that we don't want.
    if next_char in g.numeric_start_chars and token.is_datetime():
        return True

    return False


def lexer(s: str, g=PVLGrammar(), d=PVLDecoder()):
    """This is a generator function that returns pvl.Token objects
    based on the passed in string, *s*, when the generator's
    next() is called.

    A call to send(*t*) will 'return' the value *t* to the
    generator, which will be yielded upon calling next().
    This allows a user to 'peek' at the next token, but return it
    if they don't like what they see.

    *g* is expected to be an instance of pvl.grammar, and *d* an
    instance of pvl.decoder.  The lexer will perform differently,
    given different values of *g* and *d*.
    """
    c_info = _prepare_comment_tuples(g.comments)
    # print(c_info)

    lexeme = ""
    preserve = dict(state=Preserve.FALSE, end=None)
    for i, char in enumerate(s):
        if not g.char_allowed(char):
            raise LexerError(
                f'The character "{char}" (ord: {ord(char)}) '
                " is not allowed by the grammar.",
                s,
                i,
                lexeme,
            )

        prev_char = _prev_char(s, i)
        next_char = _next_char(s, i)

        # print(repr(f'lexeme at top: ->{lexeme}<-, char: {char}, '
        #            f'prev: {prev_char}, next: {next_char}, '
        #            f'{preserve}'))

        (lexeme, preserve) = lex_char(
            char, prev_char, next_char, lexeme, preserve, g, c_info
        )

        # print(repr(f'       at bot: ->{lexeme}<-,          '
        #            f'                  '
        #            f'{preserve}'))

        # Now having dealt with char, decide whether to
        # go on continue accumulating the lexeme, or yield it.

        if lexeme == "":
            continue

        try:
            # The ``while t is not None: yield None; t = yield(t)``
            # construction below allows a user of the lexer to
            # yield a token, not like what they see, and then use
            # the generator's send() function to put the token
            # back into the generator.
            #
            # The first ``yield None`` in there allows the call to
            # send() on this generator to return None, and keep the
            # value of *t* ready for the next call of next() on the
            # generator.  This is the magic that allows a user to
            # 'return' a token to the generator.
            tok = Token(lexeme, grammar=g, decoder=d, pos=firstpos(lexeme, i))

            if lex_continue(char, next_char, lexeme, tok, preserve, g):
                # Any lexeme state that we want to just allow
                # to run around again and don't want to get
                # caught by the clause in the elif, should
                # test true via lex_continue()
                continue

            elif (
                next_char is None
                or not g.char_allowed(next_char)
                or next_char in g.whitespace
                or next_char in g.reserved_characters
                or s.startswith(tuple(p[0] for p in g.comments), i + 1)
                or lexeme.endswith(tuple(p[1] for p in g.comments))
                or lexeme in g.reserved_characters
                or tok.is_quoted_string()
            ):
                # print(f'yielding {tok}')
                t = yield tok
                while t is not None:
                    yield None
                    t = yield t
                lexeme = ""
            else:
                continue

        except ValueError as err:
            raise LexerError(err, s, i, lexeme)



# -*- coding: utf-8 -*-
"""Describes the language aspects of PVL dialects.

These grammar objects are not particularly meant to be easily
user-modifiable during running of an external program, which is why
they have no arguments at initiation time, nor are there any methods
or functions to modify them.  This is because these grammar objects
are used both for reading and writing PVL-text.  As such, objects
like PVLGrammar and ODLGrammar shouldn't be altered, because if
they are, then the PVL-text written out with them wouldn't conform
to the spec.

Certainly, these objects do have attributes that can be altered,
but unless you've carefully read the code, it isn't recommended.

Maybe someday we'll add a more user-friendly interface to allow that,
but in the meantime, just leave an Issue on the GitHub repo.
"""

# Copyright 2019-2021, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import re
from collections import abc
from datetime import timezone


class PVLGrammar:
    """Describes a PVL grammar for use by the lexer and parser.

    The reference for this grammar is the CCSDS-641.0-B-2 'Blue Book'.
    """

    spacing_characters = (" ", "\t")
    format_effectors = ("\n", "\r", "\v", "\f")

    # Tuple of characters to be recognized as PVL White Space
    # (used to separate syntactic elements and promote readability,
    # but the amount or presence of White Space may not be used to
    # provide different meanings).
    whitespace = spacing_characters + format_effectors

    # Tuple of characters that may not occur in Parameter Names,
    # Unquoted Strings, nor Block Names.
    reserved_characters = (
        "&",
        "<",
        ">",
        "'",
        "{",
        "}",
        ",",
        "[",
        "]",
        "=",
        "!",
        "#",
        "(",
        ")",
        "%",
        "+",
        '"',
        ";",
        "~",
        "|",
    )

    # If there are any reserved_characters that might start a number,
    # they need to be added to numeric_start_chars, otherwise that
    # character will get lexed separately from the rest.
    # Technically, since '-' isn't in reserved_characters, it isn't needed,
    # but it doesn't hurt to keep it here.
    numeric_start_chars = ("+", "-")

    delimiters = (";",)

    # Tuple of two-tuples with each two-tuple containing a pair of character
    # sequences that enclose a comment.
    comments = (("/*", "*/"),)

    # A note on keywords: they should always be compared with
    # the str.casefold() function.
    # So 'NULL'.casefold(), 'Null'.casefold(), and 'NuLl".casefold()
    # all compare equals to none_keyword.casefold().
    none_keyword = "NULL"
    true_keyword = "TRUE"
    false_keyword = "FALSE"
    group_pref_keywords = ("BEGIN_GROUP", "END_GROUP")
    group_keywords = {"GROUP": "END_GROUP", "BEGIN_GROUP": "END_GROUP"}
    object_pref_keywords = ("BEGIN_OBJECT", "END_OBJECT")
    object_keywords = {"OBJECT": "END_OBJECT", "BEGIN_OBJECT": "END_OBJECT"}
    aggregation_keywords = dict()
    aggregation_keywords.update(group_keywords)
    aggregation_keywords.update(object_keywords)
    end_statements = ("END",)
    reserved_keywords = set(end_statements)
    for p in aggregation_keywords.items():
        reserved_keywords |= set(p)

    quotes = ('"', "'")
    set_delimiters = ("{", "}")
    sequence_delimiters = ("(", ")")
    units_delimiters = ("<", ">")

    # [sign]radix#non_decimal_integer#
    _s = r"(?P<sign>[+-]?)"
    nondecimal_pre_re = re.compile(fr"{_s}(?P<radix>2|8|16)#")
    binary_re = re.compile(fr"{_s}(?P<radix>2)#(?P<non_decimal>[01]+)#")
    octal_re = re.compile(fr"{_s}(?P<radix>8)#(?P<non_decimal>[0-7]+)#")
    hex_re = re.compile(fr"{_s}(?P<radix>16)#(?P<non_decimal>[0-9A-Fa-f]+)#")
    nondecimal_re = re.compile(
        fr"{nondecimal_pre_re.pattern}(?P<non_decimal>[0-9|A-Fa-f]+)#"
    )

    # The PVL Blue Book says that all PVl Date/Time Values are represented
    # in Universal Coordinated Time
    default_timezone = timezone.utc
    _d_formats = ("%Y-%m-%d", "%Y-%j")
    _t_formats = ("%H:%M", "%H:%M:%S", "%H:%M:%S.%f")
    date_formats = _d_formats + tuple(x + "Z" for x in _d_formats)
    time_formats = _t_formats + tuple(x + "Z" for x in _t_formats)
    datetime_formats = list()
    for d in _d_formats:
        for t in _t_formats:
            datetime_formats.append(f"{d}T{t}")
            datetime_formats.append(f"{d}T{t}Z")

    # I really didn't want to write these, because it is so easy to
    # make a mistake with time regexes, but they're they only way
    # to parse times with 60 seconds in them.  The above regexes and
    # the datetime library are used for all other time parsing.
    _H_frag = r"(?P<hour>0\d|1\d|2[0-3])"  # 00 to 23
    _M_frag = r"(?P<minute>[0-5]\d)"  # 00 to 59
    _f_frag = r"(\.(?P<microsecond>\d+))"  # 1 or more digits
    _Y_frag = r"(?P<year>\d{3}[1-9])"  # 0001 to 9999
    _m_frag = r"(?P<month>0[1-9]|1[0-2])"  # 01 to 12
    _d_frag = r"(?P<day>0[1-9]|[12]\d|3[01])"  # 01 to 31
    _Ymd_frag = fr"{_Y_frag}-{_m_frag}-{_d_frag}"
    # 001 to 366:
    _j_frag = r"(?P<doy>(00[1-9]|0[1-9]\d)|[12]\d{2}|3[0-5]\d|36[0-6])"
    _Yj_frag = fr"{_Y_frag}-{_j_frag}"
    _time_frag = fr"{_H_frag}:{_M_frag}:60{_f_frag}?Z?"  # Only times with 60 s
    # _time_frag = fr'{_H_frag}:{_M_frag}]'  # Only times with 60 s
    leap_second_Ymd_re = re.compile(fr"({_Ymd_frag}T)?{_time_frag}")
    leap_second_Yj_re = re.compile(fr"({_Yj_frag}T)?{_time_frag}")

    def char_allowed(self, char):
        """Returns true if *char* is allowed in the PVL Character Set.

        This is defined as most of the ISO 8859-1 'latin-1' character
        set with some exclusions.
        """
        if len(char) != 1:
            raise Exception

        o = ord(char)

        # The vertical tab, ord('\t') = 11, is mistakenly
        # shaded on page B-3 of the PVL specification.
        if (
            o > 255
            or (0 <= o <= 8)
            or
            # o == 11 or
            (14 <= o <= 31)
            or (127 <= o <= 159)
        ):
            return False
        else:
            return True


class ODLGrammar(PVLGrammar):
    """This defines an ODL grammar.

    The reference for this grammar is the PDS3 Standards Reference
    (version 3.8, 27 Feb 2009) Chapter 12: Object Description
    Language Specification and Usage.
    """

    group_pref_keywords = ("GROUP", "END_GROUP")
    object_pref_keywords = ("OBJECT", "END_OBJECT")

    # ODL does not allow times with a seconds value of 60.
    leap_second_Ymd_re = None
    leap_second_Yj_re = None

    # ODL allows "local" times without a timezone specifier.
    default_timezone = None

    # ODL allows the radix to be from 2 to 16, but the optional sign
    # must be after the first octothorpe (#).  Why ODL thought this was
    # an important difference to make from PVL, I have no idea.
    # radix#[sign]non_decimal_integer#
    nondecimal_pre_re = re.compile(fr"(?P<radix>[2-9]|1[0-6])#{PVLGrammar._s}")
    nondecimal_re = re.compile(
        fr"{nondecimal_pre_re.pattern}(?P<non_decimal>[0-9A-Fa-f]+)#"
    )

    def char_allowed(self, char):
        """Returns true if *char* is allowed in the ODL Character Set.

        The ODL Character Set is limited to ASCII.  This is fewer
        characters than PVL, but appears to allow more control
        characters to be in quoted strings than PVL does.
        """
        if len(char) != 1:
            raise Exception

        try:
            char.encode(encoding="ascii")
            return True
        except UnicodeError:
            return False


class PDSGrammar(ODLGrammar):
    """This defines a PDS3 ODL grammar.

    The reference for this grammar is the PDS3 Standards Reference
    (version 3.8, 27 Feb 2009) Chapter 12: Object Description
    Language Specification and Usage.
    """

    # The PDS spec only allows allows miliseconds, not microseconds,
    # but there is only a %f microseconds time format specifier in
    # Python, and no miliseconds format specifier, so dealing with
    # only the miliseconds will have to be enforced at the encoder and
    # decoder stages.

    # PDSLabels default to UTC:
    default_timezone = timezone.utc


class ISISGrammar(PVLGrammar):
    """This defines the ISIS version of PVL.

       This is valid as of ISIS 3.9, and before, at least.

       The ISIS 'Pvl' object typically writes out parameter
       values and keywords in CamelCase (e.g. 'Group', 'End_Group',
       'CenterLatitude', etc.), but it will accept all-uppercase
       versions.

       Technically, since the ISIS 'Pvl' object which parses
       PVL text into C++ objects for ISIS programs to work with
       does not recognize the 'BEGIN_<GROUP|OBJECT>' construction,
       this means that ISIS does not parse PVL text that would be
       valid according to the PVL, ODL, or PDS3 specs.
    """

    # The other thing that ISIS seems to be doing differently is to
    # split any text of all kinds with a dash continuation character.  This
    # is currently handled in the OmniParser.parse() function.

    # At
    # https://astrodiscuss.usgs.gov/t/what-pvl-specification-does-isis-conform-to/
    #
    # Stuart Sides, ISIS developer, says:
    #     The ISIS3 implementation of PVL/ODL (like) does not strictly
    #     follow any of the published standards. It was based on PDS3
    #     ODL from the 1990s, but has several extensions (your example
    #     of continuation lines) adopted from existing and prior data
    #     sets from ISIS2, PDS, JAXA, ISRO, ..., and extensions used
    #     only within ISIS3 files (cub, net). This is one of the
    #     reasons using ISIS cube files as an archive format has been
    #     strongly discouraged. So to answer your question, there is
    #     no published specification for ISIS3 PVL.

    # The ISIS parser (at least <=3.9) doesn't recognize the
    # 'BEGIN_<GROUP|OBJECT>' construction, which is why we must
    # have a separate grammar object.  Since we're at it, we might
    # as well use the *_pref_keywords to indicate the CamelCase
    # that ISIS folks are expecting.
    group_pref_keywords = ("Group", "End_Group")
    group_keywords = {"GROUP": "END_GROUP"}
    object_pref_keywords = ("Object", "End_Object")
    object_keywords = {"OBJECT": "END_OBJECT"}

    # A single-line comment that starts with the octothorpe (#) is not part
    # of PVL or ODL, but it is used when ISIS writes out comments.
    comments = (("/*", "*/"), ("#", "\n"))

    def __init__(self):
        # ISIS allows for + characters in Unquoted String values.
        self.reserved_characters = tuple(
            self.adjust_reserved_characters(self.reserved_characters)
        )

    @staticmethod
    def adjust_reserved_characters(chars: abc.Iterable):
        # ISIS allows for + characters in Unquoted String values.
        # Removing the plus from the reserved characters allows for
        # that, but might lead to other parsing errors, so be on the lookout.
        rc = list(chars)
        rc.remove("+")
        return rc


class OmniGrammar(PVLGrammar):
    """A broadly permissive grammar.

    This grammar does not follow a specification, but is meant to allow
    the broadest possible ingestion of PVL-like text that is found.

    This grammar should not be used to write out Python objects to PVL,
    instead please use one of the grammars that follows a published
    specification, like the PVLGrammar or the ODLGrammar.
    """

    # Interestingly, a single-line comment that starts with the
    # octothorpe (#) is neither part of PVL nor ODL, but people use
    # it all the time.
    comments = (("/*", "*/"), ("#", "\n"))

    # ODL allows the radix to be from 2 to 16, and allows the sign to be
    # 'inside' the octothorpes, so we need to allow for the wide variety
    # of radix, and the variational placement of the optional sign:
    # [sign]radix#[sign]non_decimal_integer#
    _ss = r"(?P<second_sign>[+-]?)"
    nondecimal_pre_re = re.compile(
        PVLGrammar._s + fr"(?P<radix>[2-9]|1[0-6])#{_ss}"
    )
    nondecimal_re = re.compile(
        nondecimal_pre_re.pattern + r"(?P<non_decimal>[0-9A-Fa-f]+)#"
    )

    def __init__(self):
        # Handle the fact that ISIS writes out unquoted plus signs.
        # See ISISGrammar for details.
        self.reserved_characters = tuple(
            ISISGrammar.adjust_reserved_characters(self.reserved_characters)
        )



#!/usr/bin/env python
"""This module has tests for the pvl decoder functions."""

# Copyright 2019-2021, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import itertools
import unittest
from decimal import Decimal

from pvl.decoder import PVLDecoder, ODLDecoder, PDSLabelDecoder, for_try_except
from pvl.collections import Quantity


class TestForTryExcept(unittest.TestCase):
    def test_for_try_except(self):
        self.assertEqual(
            5, for_try_except(ValueError, int, ("frank", "7.7", "5"))
        )
        self.assertRaises(
            ValueError, for_try_except, ValueError, int, ("frank", "7.7", "a")
        )

        self.assertEqual(
            datetime.date(2001, 1, 1),
            for_try_except(
                ValueError,
                datetime.datetime.strptime,
                itertools.repeat("2001-001"),
                ("%Y-%m-%d", "%Y-%j"),
            ).date(),
        )


class TestDecoder(unittest.TestCase):
    def setUp(self):
        self.d = PVLDecoder()

    def test_decode_quoted_string(self):
        self.assertEqual("Quoted", self.d.decode_quoted_string('"Quoted"'))
        self.assertEqual(
            'He said, "hello"',
            self.d.decode_quoted_string("""'He said, "hello"'"""),
        )
        self.assertEqual(
            'She said, \\"bye\\"',
            self.d.decode_quoted_string(r"'She said, \"bye\"'"),
        )
        self.assertEqual(
            "No\\tin Python", self.d.decode_quoted_string(r"'No\tin Python'")
        )
        self.assertEqual(
            "Line -\n Continued",
            self.d.decode_quoted_string("'Line -\n Continued'"),
        )

        # print(self.d.decode_quoted_string("""'mixed"\\'quotes'"""))

    def test_decode_unquoted_string(self):
        self.assertEqual("Unquoted", self.d.decode_unquoted_string("Unquoted"))

        for s in (
            'hhhhh"hello"',
            "Reserved=",
            "No\tin Python",
            "Line -\n Continued",
        ):
            with self.subTest(string=s):
                self.assertRaises(ValueError, self.d.decode_unquoted_string, s)

    def test_decode_decimal(self):
        for p in (
            ("125", 125),
            ("+211109", 211109),
            ("-79", -79),
            ("69.35", 69.35),
            ("+12456.345", 12456.345),  # Integers
            ("-0.23456", -0.23456),
            (".05", 0.05),
            ("-7.", -7),  # Floating
            ("-2.345678E12", -2345678000000.0),
            ("1.567E-10", 1.567e-10),
            ("+4.99E+3", 4990.0),
        ):  # Exponential
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.d.decode_decimal(p[0]))

        for s in ("2#0101#", "frank"):
            with self.subTest(string=s):
                self.assertRaises(ValueError, self.d.decode_decimal, s)

    def test_decode_withDecimal(self):
        d = PVLDecoder(real_cls=Decimal)
        s = "123.450"
        self.assertEqual(d.decode_decimal(s), Decimal(s))

        self.assertRaises(ValueError, d.decode_decimal, "fruit")

    def test_decode_non_decimal(self):
        for p in (
            ("2#0101#", 5),
            ("+2#0101#", 5),
            ("-2#0101#", -5),  # Binary
            ("8#0107#", 71),
            ("+8#0156#", 110),
            ("-8#0134#", -92),  # Oct
            ("16#100A#", 4106),
            ("+16#23Bc#", 9148),
            ("-16#98ef#", -39151),
        ):  # Hex
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.d.decode_non_decimal(p[0]))

    def test_decode_datetime(self):
        utc = datetime.timezone.utc
        for p in (
            ("2001-01-01", datetime.date(2001, 1, 1)),
            ("2001-027", datetime.date(2001, 1, 27)),
            ("2001-027Z", datetime.date(2001, 1, 27)),
            ("23:45", datetime.time(23, 45, tzinfo=utc)),
            ("01:42:57", datetime.time(1, 42, 57, tzinfo=utc)),
            ("12:34:56.789", datetime.time(12, 34, 56, 789000, tzinfo=utc)),
            (
                "2001-027T23:45",
                datetime.datetime(2001, 1, 27, 23, 45, tzinfo=utc),
            ),
            (
                "2001-01-01T01:34Z",
                datetime.datetime(2001, 1, 1, 1, 34, tzinfo=utc),
            ),
            ("01:42:57Z", datetime.time(1, 42, 57, tzinfo=utc)),
            ("2001-12-31T01:59:60.123Z", "2001-12-31T01:59:60.123Z"),
            ("2001-12-31T01:59:60.123456789", "2001-12-31T01:59:60.123456789"),
            ("01:00:60", "01:00:60"),
        ):
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.d.decode_datetime(p[0]))

        self.assertRaises(ValueError, self.d.decode_datetime, "frank")

        fancy = "2001-001T01:10:39+7"
        self.assertRaises(ValueError, self.d.decode_datetime, fancy)

    def test_decode_simple_value(self):
        for p in (
            ("2001-01-01", datetime.date(2001, 1, 1)),
            ("2#0101#", 5),
            ("-79", -79),
            ("Unquoted", "Unquoted"),
            ('"Quoted"', "Quoted"),
            ("Null", None),
            ("TRUE", True),
            ("false", False),
        ):
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.d.decode_simple_value(p[0]))

    def test_decode_quantity(self):
        q = self.d.decode_quantity("15", "m/s")
        self.assertEqual(q, Quantity("15", "m/s"))

        try:
            from astropy import units as u

            d = PVLDecoder(quantity_cls=u.Quantity)
            q = d.decode_quantity("15", "m/s")
            self.assertEqual(q, u.Quantity("15", "m/s"))
        except ImportError:  # astropy isn't available.
            pass

        try:
            from pint import Quantity as pintquant

            d = PVLDecoder(quantity_cls=pintquant)
            q = d.decode_quantity("15", "m/s")
            self.assertEqual(q, pintquant("15", "m/s"))
        except ImportError:  # pint isn't available.
            pass


class TestODLDecoder(unittest.TestCase):
    def setUp(self):
        self.d = ODLDecoder()

    def test_decode_datetime(self):
        utc = datetime.timezone.utc

        for p in (
                ("1990-07-04", datetime.date(1990, 7, 4)),
                ("1990-158", datetime.date(1990, 6, 7)),
                ("2001-001", datetime.date(2001, 1, 1)),
                ("2001-01-01", datetime.date(2001, 1, 1)),
                ("12:00", datetime.time(12)),
                ("12:00:45", datetime.time(12, 0, 45)),
                (
                        "12:00:45.4571",
                        datetime.time(12, 0, 45, 457100),
                ),
                ("15:24:12Z", datetime.time(15, 24, 12, tzinfo=utc)),
                ("1990-07-04T12:00", datetime.datetime(1990, 7, 4, 12)),
                (
                    "1990-158T15:24:12Z",
                    datetime.datetime(1990, 6, 7, 15, 24, 12, tzinfo=utc),
                ),
        ):
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.d.decode_datetime(p[0]))

        self.assertRaises(ValueError, self.d.decode_datetime, "01:00:60")

        try:
            from dateutil import tz
            tz_plus_7 = tz.tzoffset("+7", datetime.timedelta(hours=7))

            for p in (
                ("01:12:22+07", datetime.time(1, 12, 22, tzinfo=tz_plus_7)),
                ("01:12:22+7", datetime.time(1, 12, 22, tzinfo=tz_plus_7)),
                (
                    "01:10:39.4575+07",
                    datetime.time(1, 10, 39, 457500, tzinfo=tz_plus_7),
                ),
                (
                    "2001-001T01:10:39+7",
                    datetime.datetime(2001, 1, 1, 1, 10, 39, tzinfo=tz_plus_7),
                ),
                (
                    "2001-001T01:10:39.457591+7",
                    datetime.datetime(
                        2001, 1, 1, 1, 10, 39, 457591, tzinfo=tz_plus_7
                    ),
                ),
            ):
                with self.subTest(pair=p):
                    self.assertEqual(p[1], self.d.decode_datetime(p[0]))

        except ImportError:  # dateutil isn't available.
            pass


class TestPDS3Decoder(unittest.TestCase):
    def setUp(self):
        self.d = PDSLabelDecoder()

    def test_decode_datetime(self):
        utc = datetime.timezone.utc

        for p in (
            ("1990-07-04", datetime.date(1990, 7, 4)),
            ("1990-158", datetime.date(1990, 6, 7)),
            ("2001-001", datetime.date(2001, 1, 1)),
            ("2001-01-01", datetime.date(2001, 1, 1)),
            ("12:00", datetime.time(12, tzinfo=utc)),
            ("12:00:45", datetime.time(12, 0, 45, tzinfo=utc)),
            ("15:24:12Z", datetime.time(15, 24, 12, tzinfo=utc)),
            (
                "1990-158T15:24:12Z",
                datetime.datetime(1990, 6, 7, 15, 24, 12, tzinfo=utc)),
            (
                "12:00:45.457",
                datetime.time(12, 0, 45, 457000, tzinfo=utc),
            ),
            (
                "1990-07-04T12:00",
                datetime.datetime(1990, 7, 4, 12, tzinfo=utc),
            ),
        ):
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.d.decode_datetime(p[0]))

        for t in (
                "01:12:22+07",
                "01:12:22+7",
                "01:10:39.4575+07",
                "2001-001T01:10:39+7",
                "2001-001T01:10:39.457591+7",
                "2001-001T01:10:39.457591",
        ):
            with self.subTest(time=t):
                self.assertRaises(ValueError, self.d.decode_datetime, t)



# -*- coding: utf-8 -*-
"""Parameter Value Langage encoder.

An encoder deals with converting Python objects into
string values that conform to a PVL specification.
"""

# Copyright 2015, 2019-2021, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import datetime
import re
import textwrap

from collections import abc, namedtuple
from decimal import Decimal
from warnings import warn

from .collections import PVLObject, PVLGroup, Quantity
from .grammar import PVLGrammar, ODLGrammar, PDSGrammar, ISISGrammar
from .token import Token
from .decoder import PVLDecoder, ODLDecoder, PDSLabelDecoder


class QuantTup(namedtuple("QuantTup", ["cls", "value_prop", "units_prop"])):
    """
    This class is just a convenient namedtuple for internally keeping track
    of quantity classes that encoders can deal with.  In general, users
    should not be instantiating this, instead use your encoder's
    add_quantity_cls() function.
    """


class PVLEncoder(object):
    """An encoder based on the rules in the CCSDS-641.0-B-2 'Blue Book'
    which defines the PVL language.

    :param grammar: A pvl.grammar object, if None or not specified, it will
                    be set to the grammar parameter of *decoder* (if
                    *decoder* is not None) or will default to PVLGrammar().
    :param grammar: defaults to pvl.grammar.PVLGrammar().
    :param decoder: defaults to pvl.decoder.PVLDecoder().
    :param indent: specifies the number of spaces that will be used to
        indent each level of the PVL document, when Groups or Objects
        are encountered, defaults to 2.
    :param width: specifies the number of characters in width that each
        line should have, defaults to 80.
    :param aggregation_end: when True the encoder will print the value
        of the aggregation's Block Name in the End Aggregation Statement
        (e.g. END_GROUP = foo), and when false, it won't (e.g. END_GROUP).
        Defaults to True.
    :param end_delimiter: when True the encoder will print the grammar's
        delimiter (e.g. ';' for PVL) after each statement, when False
        it won't.  Defaults to True.
    :param newline: is the string that will be placed at the end of each
        'line' of output (and counts against *width*), defaults to '\\\\n'.
    :param group_class: must this class will be tested against with
        isinstance() to determine if various elements of the dict-like
        passed to encode() should be encoded as a PVL Group or PVL Object,
        defaults to PVLGroup.
    :param object_class: must be a class that can take a *group_class*
        object in its constructor (essentially converting a *group_class*
        to an *object_class*), otherwise will raise TypeError.  Defaults
        to PVLObject.
    """

    def __init__(
        self,
        grammar=None,
        decoder=None,
        indent: int = 2,
        width: int = 80,
        aggregation_end: bool = True,
        end_delimiter: bool = True,
        newline: str = "\n",
        group_class=PVLGroup,
        object_class=PVLObject,
    ):

        if grammar is None:
            if decoder is not None:
                self.grammar = decoder.grammar
            else:
                self.grammar = PVLGrammar()
        elif isinstance(grammar, PVLGrammar):
            self.grammar = grammar
        else:
            raise Exception

        if decoder is None:
            self.decoder = PVLDecoder(self.grammar)
        elif isinstance(decoder, PVLDecoder):
            self.decoder = decoder
        else:
            raise Exception

        self.indent = indent
        self.width = width
        self.end_delimiter = end_delimiter
        self.aggregation_end = aggregation_end
        self.newline = newline

        # This list of 3-tuples *always* has our own pvl quantity object,
        # and should *only* be added to with self.add_quantity_cls().
        self.quantities = [QuantTup(Quantity, "value", "units")]
        self._import_quantities()

        if issubclass(group_class, abc.Mapping):
            self.grpcls = group_class
        else:
            raise TypeError("The group_class must be a Mapping type.")

        if issubclass(object_class, abc.Mapping):
            self.objcls = object_class
        else:
            raise TypeError("The object_class must be a Mapping type.")

        try:
            self.objcls(self.grpcls())
        except TypeError:
            raise TypeError(
                f"The object_class type ({object_class}) cannot be "
                f"instantiated with an argument that is of type "
                f"group_class ({group_class})."
            )

        # Finally, let's keep track of everything we consider "numerical":
        self.numeric_types = (int, float, self.decoder.real_cls, Decimal)

    def _import_quantities(self):
        warn_str = (
            "The {} library is not present, so {} objects will "
            "not be properly encoded."
        )
        try:
            from astropy import units as u

            self.add_quantity_cls(u.Quantity, "value", "unit")
        except ImportError:
            warn(
                warn_str.format("astropy", "astropy.units.Quantity"),
                ImportWarning,
            )

        try:
            from pint import Quantity as q

            self.add_quantity_cls(q, "magnitude", "units")
        except ImportError:
            warn(warn_str.format("pint", "pint.Quantity"), ImportWarning)

    def add_quantity_cls(self, cls, value_prop: str, units_prop: str):
        """Adds a quantity class to the list of possible
        quantities that this encoder can handle.

        :param cls: The name of a quantity class that can be tested
            with ``isinstance()``.
        :param value_prop: A string that is the property name of
            *cls* that contains the value or magnitude of the quantity
            object.
        :param units_prop: A string that is the property name of
            *cls* that contains the units element of the quantity
            object.
        """
        if not isinstance(cls, type):
            raise TypeError(f"The cls given ({cls}) is not a Python class.")

        # If a quantity object can't encode "one meter" its probably not
        # going to work for us.
        test_cls = cls(1, "m")
        for prop in (value_prop, units_prop):
            if not hasattr(test_cls, prop):
                raise AttributeError(
                    f"The class ({cls}) does not have an "
                    f" attribute named {prop}."
                )

        self.quantities.append(QuantTup(cls, value_prop, units_prop))

    def format(self, s: str, level: int = 0) -> str:
        """Returns a string derived from *s*, which
        has leading space characters equal to
        *level* times the number of spaces specified
        by this encoder's indent property.

        It uses the textwrap library to wrap long lines.
        """

        prefix = level * (self.indent * " ")

        if len(prefix + s + self.newline) > self.width and "=" in s:
            (preq, _, posteq) = s.partition("=")
            new_prefix = prefix + preq.strip() + " = "

            lines = textwrap.wrap(
                posteq.strip(),
                width=(self.width - len(self.newline)),
                replace_whitespace=False,
                initial_indent=new_prefix,
                subsequent_indent=(" " * len(new_prefix)),
                break_long_words=False,
                break_on_hyphens=False,
            )
            return self.newline.join(lines)
        else:
            return prefix + s

    def encode(self, module: abc.Mapping) -> str:
        """Returns a ``str`` formatted as a PVL document based
        on the dict-like *module* object
        according to the rules of this encoder.
        """
        lines = list()
        lines.append(self.encode_module(module, 0))

        end_line = self.grammar.end_statements[0]
        if self.end_delimiter:
            end_line += self.grammar.delimiters[0]

        lines.append(end_line)

        # Final check to ensure we're sending out the right character set:
        s = self.newline.join(lines)

        for i, c in enumerate(s):
            if not self.grammar.char_allowed(c):
                raise ValueError(
                    "Encountered a character that was not "
                    "a valid character according to the "
                    'grammar: "{}", it is in: '
                    '"{}"'.format(c, s[i - 5, i + 5])
                )

        return self.newline.join(lines)

    def encode_module(self, module: abc.Mapping, level: int = 0) -> str:
        """Returns a ``str`` formatted as a PVL module based
        on the dict-like *module* object according to the
        rules of this encoder, with an indentation level
        of *level*.
        """
        lines = list()

        # To align things on the equals sign, just need to normalize
        # the non-aggregation key length:

        non_agg_key_lengths = list()
        for k, v in module.items():
            if not isinstance(v, abc.Mapping):
                non_agg_key_lengths.append(len(k))
        longest_key_len = max(non_agg_key_lengths, default=0)

        for k, v in module.items():
            if isinstance(v, abc.Mapping):
                lines.append(self.encode_aggregation_block(k, v, level))
            else:
                lines.append(
                    self.encode_assignment(k, v, level, longest_key_len)
                )
        return self.newline.join(lines)

    def encode_aggregation_block(
        self, key: str, value: abc.Mapping, level: int = 0
    ) -> str:
        """Returns a ``str`` formatted as a PVL Aggregation Block with
        *key* as its name, and its contents based on the
        dict-like *value* object according to the
        rules of this encoder, with an indentation level
        of *level*.
        """
        lines = list()

        if isinstance(value, self.grpcls):
            agg_keywords = self.grammar.group_pref_keywords
        elif isinstance(value, abc.Mapping):
            agg_keywords = self.grammar.object_pref_keywords
        else:
            raise ValueError("The value {value} is not dict-like.")

        agg_begin = "{} = {}".format(agg_keywords[0], key)
        if self.end_delimiter:
            agg_begin += self.grammar.delimiters[0]
        lines.append(self.format(agg_begin, level))

        lines.append(self.encode_module(value, (level + 1)))

        agg_end = ""
        if self.aggregation_end:
            agg_end += "{} = {}".format(agg_keywords[1], key)
        else:
            agg_end += agg_keywords[1]
        if self.end_delimiter:
            agg_end += self.grammar.delimiters[0]
        lines.append(self.format(agg_end, level))

        return self.newline.join(lines)

    def encode_assignment(
        self, key: str, value, level: int = 0, key_len: int = None
    ) -> str:
        """Returns a ``str`` formatted as a PVL Assignment Statement
        with *key* as its Parameter Name, and its value based
        on *value* object according to the rules of this encoder,
        with an indentation level of *level*.  It also allows for
        an optional *key_len* which indicates the width in characters
        that the Assignment Statement should be set to, defaults to
        the width of *key*.
        """
        if key_len is None:
            key_len = len(key)

        s = ""
        s += "{} = ".format(key.ljust(key_len))

        enc_val = self.encode_value(value)

        if enc_val.startswith(self.grammar.quotes):
            # deal with quoted lines that need to preserve
            # newlines
            s = self.format(s, level)
            s += enc_val

            if self.end_delimiter:
                s += self.grammar.delimiters[0]

            return s
        else:
            s += enc_val

            if self.end_delimiter:
                s += self.grammar.delimiters[0]

            return self.format(s, level)

    def encode_value(self, value) -> str:
        """Returns a ``str`` formatted as a PVL Value based
        on the *value* object according to the rules of this encoder.
        """
        try:
            return self.encode_quantity(value)
        except ValueError:
            return self.encode_simple_value(value)

    def encode_quantity(self, value) -> str:
        """Returns a ``str`` formatted as a PVL Value followed by
        a PVL Units Expression if the *value* object can be
        encoded this way, otherwise raise ValueError."""
        for (cls, v_prop, u_prop) in self.quantities:
            if isinstance(value, cls):
                return self.encode_value_units(
                    getattr(value, v_prop), getattr(value, u_prop)
                )

        raise ValueError(
            f"The value object {value} could not be "
            "encoded as a PVL Value followed by a PVL "
            f"Units Expression, it is of type {type(value)}"
        )

    def encode_value_units(self, value, units) -> str:
        """Returns a ``str`` formatted as a PVL Value from *value*
        followed by a PVL Units Expressions from *units*."""
        value_str = self.encode_simple_value(value)
        units_str = self.encode_units(str(units))
        return f"{value_str} {units_str}"

    def encode_simple_value(self, value) -> str:
        """Returns a ``str`` formatted as a PVL Simple Value based
        on the *value* object according to the rules of this encoder.
        """
        if value is None:
            return self.grammar.none_keyword
        elif isinstance(value, (set, frozenset)):
            return self.encode_set(value)
        elif isinstance(value, list):
            return self.encode_sequence(value)
        elif isinstance(
            value, (datetime.datetime, datetime.date, datetime.time)
        ):
            return self.encode_datetype(value)
        elif isinstance(value, bool):
            if value:
                return self.grammar.true_keyword
            else:
                return self.grammar.false_keyword
        elif isinstance(value, self.numeric_types):
            return str(value)
        elif isinstance(value, str):
            return self.encode_string(value)
        else:
            raise TypeError(f"{value!r} is not serializable.")

    def encode_setseq(self, values: abc.Collection) -> str:
        """This function provides shared functionality for
        encode_sequence() and encode_set().
        """
        return ", ".join([self.encode_value(v) for v in values])

    def encode_sequence(self, value: abc.Sequence) -> str:
        """Returns a ``str`` formatted as a PVL Sequence based
        on the *value* object according to the rules of this encoder.
        """
        return "(" + self.encode_setseq(value) + ")"

    def encode_set(self, value: abc.Set) -> str:
        """Returns a ``str`` formatted as a PVL Set based
        on the *value* object according to the rules of this encoder.
        """
        return "{" + self.encode_setseq(value) + "}"

    def encode_datetype(self, value) -> str:
        """Returns a ``str`` formatted as a PVL Date/Time based
        on the *value* object according to the rules of this encoder.
        If *value* is not a datetime date, time, or datetime object,
        it will raise TypeError.
        """
        if isinstance(value, datetime.datetime):
            return self.encode_datetime(value)
        elif isinstance(value, datetime.date):
            return self.encode_date(value)
        elif isinstance(value, datetime.time):
            return self.encode_time(value)
        else:
            raise TypeError(f"{value!r} is not a datetime type.")

    @staticmethod
    def encode_date(value: datetime.date) -> str:
        """Returns a ``str`` formatted as a PVL Date based
        on the *value* object according to the rules of this encoder.
        """
        return f"{value:%Y-%m-%d}"

    @staticmethod
    def encode_time(value: datetime.time) -> str:
        """Returns a ``str`` formatted as a PVL Time based
        on the *value* object according to the rules of this encoder.
        """
        s = f"{value:%H:%M}"

        if value.microsecond:
            s += f":{value:%S.%f}"
        elif value.second:
            s += f":{value:%S}"

        return s

    def encode_datetime(self, value: datetime.datetime) -> str:
        """Returns a ``str`` formatted as a PVL Date/Time based
        on the *value* object according to the rules of this encoder.
        """
        date = self.encode_date(value)
        time = self.encode_time(value)
        return date + "T" + time

    def needs_quotes(self, s: str) -> bool:
        """Returns true if *s* must be quoted according to this
        encoder's grammar, false otherwise.
        """
        if any(c in self.grammar.whitespace for c in s):
            return True

        if s in self.grammar.reserved_keywords:
            return True

        tok = Token(s, grammar=self.grammar, decoder=self.decoder)
        return not tok.is_unquoted_string()

    def encode_string(self, value) -> str:
        """Returns a ``str`` formatted as a PVL String based
        on the *value* object according to the rules of this encoder.
        """
        s = str(value)

        if self.needs_quotes(s):
            for q in self.grammar.quotes:
                if q not in s:
                    return q + s + q
            else:
                raise ValueError(
                    "All of the quote characters, "
                    f"{self.grammar.quotes}, were in the "
                    f'string ("{s}"), so it could not be quoted.'
                )
        else:
            return s

    def encode_units(self, value: str) -> str:
        """Returns a ``str`` formatted as a PVL Units Value based
        on the *value* object according to the rules of this encoder.
        """
        return (
            self.grammar.units_delimiters[0]
            + value
            + self.grammar.units_delimiters[1]
        )


class ODLEncoder(PVLEncoder):
    """An encoder based on the rules in the PDS3 Standards Reference
    (version 3.8, 27 Feb 2009) Chapter 12: Object Description
    Language Specification and Usage for ODL only.  This is
    almost certainly not what you want.  There are very rarely
    cases where you'd want to use ODL that you wouldn't also want
    to use the PDS Label restrictions, so you probably really want
    the PDSLabelEncoder class, not this one.  Move along.

    It extends PVLEncoder.

    :param grammar: defaults to pvl.grammar.ODLGrammar().
    :param decoder: defaults to pvl.decoder.ODLDecoder().
    :param end_delimiter: defaults to False.
    :param newline: defaults to '\\\\r\\\\n'.
    """

    def __init__(
        self,
        grammar=None,
        decoder=None,
        indent=2,
        width=80,
        aggregation_end=True,
        end_delimiter=False,
        newline="\r\n",
    ):

        if grammar is None:
            grammar = ODLGrammar()

        if decoder is None:
            decoder = ODLDecoder(grammar)

        if not callable(getattr(decoder, "is_identifier", None)):
            raise TypeError(
                f"The decoder for an ODLEncoder() must have the "
                f"is_identifier() function, and this does not: {decoder}"
            )

        super().__init__(
            grammar,
            decoder,
            indent,
            width,
            aggregation_end,
            end_delimiter,
            newline,
        )

    def encode(self, module: abc.Mapping) -> str:
        """Extends parent function, but ODL requires that there must be
        a spacing or format character after the END statement and this
        adds the encoder's ``newline`` sequence.
        """
        s = super().encode(module)
        return s + self.newline

    def is_scalar(self, value) -> bool:
        """Returns a boolean indicating whether the *value* object
        qualifies as an ODL 'scalar_value'.

        ODL defines a 'scalar-value' as a numeric_value, a
        date_time_string, a text_string_value, or a symbol_value.

        For Python, these correspond to the following:

        * numeric_value: any of self.numeric_types, and Quantity whose value
          is one of the self.numeric_types.
        * date_time_string: datetime objects
        * text_string_value: str
        * symbol_value: str

        """
        for quant in self.quantities:
            if isinstance(value, quant.cls):
                if isinstance(
                    getattr(value, quant.value_prop), self.numeric_types
                ):
                    return True

        scalar_types = (
            *self.numeric_types,
            datetime.date,
            datetime.datetime,
            datetime.time,
            str
        )
        if isinstance(value, scalar_types):
            return True

        return False

    def is_symbol(self, value) -> bool:
        """Returns true if *value* is an ODL Symbol String, false otherwise.

        An ODL Symbol String is enclosed by single quotes
        and may not contain any of the following characters:

        1. The apostrophe, which is reserved as the symbol string delimiter.
        2. ODL Format Effectors
        3. Control characters

        This means that an ODL Symbol String is a subset of the PVL
        quoted string, and will be represented in Python as a ``str``.
        """
        if isinstance(value, str):
            if "'" in value:  # Item 1
                return False

            for fe in self.grammar.format_effectors:  # Item 2
                if fe in value:
                    return False

            if len(value) > self.width / 2:
                # This means that the string is long and it is very
                # likely to get wrapped and have carriage returns,
                # and thus "ODL Format Effectors" inserted later.
                # Unfortunately, without knowing the width of the
                # parameter term, and the current indent level, this
                # still may end up being incorrect threshhold.
                return False

            if value.isprintable() and len(value) > 0:  # Item 3
                return True
        else:
            return False

    def needs_quotes(self, s: str) -> bool:
        """Return true if *s* is an ODL Identifier, false otherwise.

        Overrides parent function.
        """
        return not self.decoder.is_identifier(s)

    def is_assignment_statement(self, s) -> bool:
        """Returns true if *s* is an ODL Assignment Statement, false otherwise.

        An ODL Assignment Statement is either an
        element_identifier or a namespace_identifier
        joined to an element_identifier with a colon.
        """
        if self.decoder.is_identifier(s):
            return True

        (ns, _, el) = s.partition(":")

        if self.decoder.is_identifier(ns) and self.decoder.is_identifier(el):
            return True

        return False

    def encode_assignment(self, key, value, level=0, key_len=None) -> str:
        """Overrides parent function by restricting the length of
        keywords and enforcing that they be ODL Identifiers
        and uppercasing their characters.
        """

        if key_len is None:
            key_len = len(key)

        if len(key) > 30:
            raise ValueError(
                "ODL keywords must be 30 characters or less "
                f"in length, this one is longer: {key}"
            )

        if (
            key.startswith("^") and self.is_assignment_statement(key[1:])
        ) or self.is_assignment_statement(key):
            ident = key.upper()
        else:
            raise ValueError(
                f'The keyword "{key}" is not a valid ODL ' "Identifier."
            )

        s = "{} = ".format(ident.ljust(key_len))
        s += self.encode_value(value)

        if self.end_delimiter:
            s += self.grammar.delimiters[0]

        return self.format(s, level)

    def encode_sequence(self, value) -> str:
        """Extends parent function, as ODL only allows one- and
        two-dimensional sequences of ODL scalar_values.
        """
        if len(value) == 0:
            raise ValueError("ODL does not allow empty Sequences.")

        for v in value:  # check the first dimension (list of elements)
            if isinstance(v, list):
                for i in v:  # check the second dimension (list of lists)
                    if isinstance(i, list):
                        # Shouldn't be lists of lists of lists.
                        raise ValueError(
                            "ODL only allows one- and two- "
                            "dimensional Sequences, but "
                            f"this has more: {value}"
                        )
                    elif not self.is_scalar(i):
                        raise ValueError(
                            "ODL only allows scalar_values "
                            f"within sequences: {v}"
                        )

            elif not self.is_scalar(v):
                raise ValueError(
                    "ODL only allows scalar_values within " f"sequences: {v}"
                )

        return super().encode_sequence(value)

    def encode_set(self, values) -> str:
        """Extends parent function, ODL only allows sets to contain
        scalar values.
        """

        if not all(map(self.is_scalar, values)):
            raise ValueError(
                f"ODL only allows scalar values in sets: {values}"
            )

        return super().encode_set(values)

    def encode_value(self, value):
        """Extends parent function by only allowing Units Expressions for
        numeric values.
        """
        for quant in self.quantities:
            if isinstance(value, quant.cls):
                if isinstance(
                    getattr(value, quant.value_prop),
                    self.numeric_types
                ):
                    return super().encode_value(value)
                else:
                    raise ValueError(
                        "Unit expressions are only allowed "
                        f"following numeric values: {value}"
                    )

        return super().encode_value(value)

    def encode_string(self, value):
        """Extends parent function by appropriately quoting Symbol
        Strings.
        """
        if self.decoder.is_identifier(value):
            return value
        elif self.is_symbol(value):
            return "'" + value + "'"
        else:
            return super().encode_string(value)

    def encode_time(self, value: datetime.time) -> str:
        """Extends parent function since ODL allows a time zone offset
        from UTC to be included, and otherwise recommends that times
        be suffixed with a 'Z' to clearly indicate that they are in UTC.
        """
        if value.tzinfo is None:
            raise ValueError(
                f"ODL cannot output local times, and this time does not "
                f"have a timezone offset: {value}"
            )

        t = super().encode_time(value)

        if value.utcoffset() == datetime.timedelta():
            return t + "Z"
        else:
            td_str = str(value.utcoffset())
            (h, m, s) = td_str.split(":")
            if s != "00":
                raise ValueError(
                    "The datetime value had a timezone offset "
                    f"with seconds values ({value}) which is "
                    "not allowed in ODL."
                )
            if m == "00":
                return t + f"+{h:0>2}"
            else:
                return t + f"+{h:0>2}:{m}"

        return t

    def encode_units(self, value) -> str:
        """Overrides parent function since ODL limits what characters
        and operators can be present in Units Expressions.
        """

        # if self.is_identifier(value.strip('*/()-')):
        if self.decoder.is_identifier(re.sub(r"[\s*/()-]", "", value)):

            if "**" in value:
                exponents = re.findall(r"\*\*.+?", value)
                for e in exponents:
                    if re.search(r"\*\*-?\d+", e) is None:
                        raise ValueError(
                            "The exponentiation operator (**) in "
                            f'this Units Expression "{value}" '
                            "is not a decimal integer."
                        )

            return (
                self.grammar.units_delimiters[0]
                + value
                + self.grammar.units_delimiters[1]
            )
        else:
            raise ValueError(
                f'The value, "{value}", does not conform to '
                "the specification for an ODL Units Expression."
            )


class PDSLabelEncoder(ODLEncoder):
    """An encoder based on the rules in the PDS3 Standards Reference
    (version 3.8, 27 Feb 2009) Chapter 12: Object Description
    Language Specification and Usage and writes out labels that
    conform to the PDS 3 standards.

    It extends ODLEncoder.

    You are not allowed to chose *end_delimiter* or *newline*
    as the parent class allows, because to be PDS-compliant,
    those are fixed choices.  However, in some cases, the PDS3
    Standards are asymmetric, allowing for a wider variety of
    PVL-text on "read" and a more narrow variety of PVL-text
    on "write".  The default values of the PDSLabelEncoder enforce
    those strict "write" rules, but if you wish to alter them,
    but still produce PVL-text that would validate against the PDS3
    standard, you may alter them.

    :param convert_group_to_object: Defaults to True, meaning that
        if a GROUP does not conform to the PDS definition of a
        GROUP, then it will be written out as an OBJECT.  If it is
        False, then an exception will be thrown if incompatible
        GROUPs are encountered. In PVL and ODL, the OBJECT and GROUP
        aggregations are interchangeable, but the PDS applies
        restrictions to what can appear in a GROUP.
    :param tab_replace: Defaults to 4 and indicates the number of
        space characters to replace horizontal tab characters with
        (since tabs aren't allowed in PDS labels).  If this is set
        to zero, tabs will not be replaced with spaces.
    :param symbol_single_quotes: Defaults to True, and if a Python `str`
        object qualifies as a PVL Symbol String, it will be written to
        PVL-text as a single-quoted string.  If False, no special
        handling is done, and any PVL Symbol String will be treated
        as a PVL Text String, which is typically enclosed with double-quotes.
    :param time_trailing_z: defaults to True, and suffixes a "Z" to
        datetimes and times written to PVL-text as the PDS encoding
        standard requires.  If False, no trailing "Z" is written.

    """

    def __init__(
        self,
        grammar=None,
        decoder=None,
        indent=2,
        width=80,
        aggregation_end=True,
        convert_group_to_object=True,
        tab_replace=4,
        symbol_single_quote=True,
        time_trailing_z=True,
    ):

        if grammar is None:
            grammar = PDSGrammar()

        if decoder is None:
            decoder = PDSLabelDecoder(grammar)

        super().__init__(
            grammar,
            decoder,
            indent,
            width,
            aggregation_end,
            end_delimiter=False,
            newline="\r\n",
        )

        self.convert_group_to_object = convert_group_to_object
        self.tab_replace = tab_replace
        self.symbol_single_quote = symbol_single_quote
        self.time_trailing_z = time_trailing_z

    def count_aggs(
        self, module: abc.Mapping, obj_count: int = 0, grp_count: int = 0
    ) -> tuple((int, int)):
        """Returns the count of OBJECT and GROUP aggregations
        that are contained within the *module* as a two-tuple
        in that order.
        """
        # This currently just counts the values in the passed
        # in module, it does not 'recurse' if those aggregations also
        # may contain aggregations.

        for k, v in module.items():
            if isinstance(v, abc.Mapping):
                if isinstance(v, self.grpcls):
                    grp_count += 1
                elif isinstance(v, self.objcls):
                    obj_count += 1
                else:
                    # We treat other dict-like Python objects as
                    # PVL Objects for the purposes of this count,
                    # because that is how they will be encoded.
                    obj_count += 1

        return obj_count, grp_count

    def encode(self, module: abc.MutableMapping) -> str:
        """Extends the parent function, by adding a restriction.
        For PDS, if there are any GROUP elements, there must be at
        least one OBJECT element in the label.  Behavior here
        depends on the value of this encoder's convert_group_to_object
        property.
        """
        (obj_count, grp_count) = self.count_aggs(module)

        if grp_count > 0 and obj_count < 1:
            if self.convert_group_to_object:
                for k, v in module.items():
                    # First try to convert any GROUPs that would not
                    # be valid PDS GROUPs.
                    if isinstance(v, self.grpcls) and not self.is_PDSgroup(v):
                        module[k] = self.objcls(v)
                        break
                else:
                    # Then just convert the first GROUP
                    for k, v in module.items():
                        if isinstance(v, self.grpcls):
                            module[k] = self.objcls(v)
                            break
                    else:
                        raise ValueError(
                            "Couldn't convert any of the GROUPs " "to OBJECTs."
                        )
            else:
                raise ValueError(
                    "This module has a GROUP element, but no "
                    "OBJECT elements, which is not allowed by "
                    "the PDS.  You could set "
                    "*convert_group_to_object* to *True* on the "
                    "encoder to try and convert a GROUP "
                    "to an OBJECT."
                )

        s = super().encode(module)
        if self.tab_replace > 0:
            return s.replace("\t", (" " * self.tab_replace))
        else:
            return s

    def is_PDSgroup(self, group: abc.Mapping) -> bool:
        """Returns true if the dict-like *group* qualifies as a PDS Group,
        false otherwise.

        PDS applies the following restrictions to GROUPS:

        1. The GROUP structure may only be used in a data product
           label which also contains one or more data OBJECT definitions.
        2. The GROUP statement must contain only attribute assignment
           statements, include pointers, or related information pointers
           (i.e., no data location pointers). If there are multiple
           values, a single statement must be used with either sequence
           or set syntax; no attribute assignment statement or pointer
           may be repeated.
        3. GROUP statements may not be nested.
        4. GROUP statements may not contain OBJECT definitions.
        5. Only PSDD elements may appear within a GROUP statement.
           *PSDD is not defined anywhere in the PDS document, so don't
           know how to test for it.*
        6. The keyword contents associated with a specific GROUP
           identifier must be identical across all labels of a single data
           set (with the exception of the “PARAMETERS” GROUP, as
           explained).

        Use of the GROUP structure must be coordinated with the
        responsible PDS discipline Node.

        Items 1 & 6 and the final sentence above, can't really be tested
        by examining a single group, but must be dealt with in a larger
        context.  The ODLEncoder.encode_module() handles #1, at least.
        You're on your own for the other two issues.

        Item 5: *PSDD* is not defined anywhere in the ODL PDS document,
        so don't know how to test for it.
        """
        (obj_count, grp_count) = self.count_aggs(group)

        # Items 3 and 4:
        if obj_count != 0 or grp_count != 0:
            return False

        # Item 2, no data location pointers:
        for k, v in group.items():
            if k.startswith("^"):
                if isinstance(v, int):
                    return False
                else:
                    for quant in self.quantities:
                        if isinstance(v, quant.cls) and isinstance(
                            getattr(v, quant.value_prop), int
                        ):
                            return False

        # Item 2, no repeated keys:
        keys = list(group.keys())
        if len(keys) != len(set(keys)):
            return False

        return True

    def encode_aggregation_block(self, key, value, level=0):
        """Extends parent function because PDS has restrictions on
        what may be in a GROUP.

        If the encoder's *convert_group_to_object* parameter is True,
        and a GROUP does not conform to the PDS definition of a GROUP,
        then it will be written out as an OBJECT.  If it is False,
        then an exception will be thrown.
        """

        # print('value at top:')
        # print(value)

        if isinstance(value, self.grpcls) and not self.is_PDSgroup(value):
            if self.convert_group_to_object:
                value = self.objcls(value)
            else:
                raise ValueError(
                    "This GROUP element is not a valid PDS "
                    "GROUP.  You could set "
                    "*convert_group_to_object* to *True* on the "
                    "encoder to try and convert the GROUP"
                    "to an OBJECT."
                )

        # print('value at bottom:')
        # print(value)

        return super().encode_aggregation_block(key, value, level)

    def encode_set(self, values) -> str:
        """Extends parent function because PDS only allows symbol values
        and integers within sets.
        """
        for v in values:
            if not self.is_symbol(v) and not isinstance(v, int):
                raise ValueError(
                    "The PDS only allows integers and symbols "
                    f"in sets: {values}"
                )

        return super().encode_set(values)

    def encode_string(self, value):
        """Extends parent function to treat Symbol Strings as Text Strings,
        which typically means that they are double-quoted and not
        single-quoted.
        """
        if self.decoder.is_identifier(value):
            return value
        elif self.is_symbol(value) and self.symbol_single_quote:
            return "'" + value + "'"
        else:
            return super(ODLEncoder, self).encode_string(value)

    def encode_time(self, value: datetime.time) -> str:
        """Overrides parent's encode_time() function because
        even though ODL allows for timezones, PDS does not.

        Not in the section on times, but at the end of the PDS
        ODL document, in section 12.7.3, para 14, it indicates that
        alternate time zones may not be used in a PDS label, only
        these:
        1. YYYY-MM-DDTHH:MM:SS.SSS.
        2. YYYY-DDDTHH:MM:SS.SSS.

        """
        s = f"{value:%H:%M}"

        if value.microsecond:
            ms = round(value.microsecond / 1000)
            if value.microsecond != ms * 1000:
                raise ValueError(
                    f"PDS labels can only encode time values to the milisecond "
                    f"precision, and this time ({value}) has too much "
                    f"precision."
                )
            else:
                s += f":{value:%S}.{ms}"
        elif value.second:
            s += f":{value:%S}"

        if (
            value.tzinfo is None or
            value.tzinfo.utcoffset(None) == datetime.timedelta(0)
        ):
            if self.time_trailing_z:
                return s + "Z"
            else:
                return s
        else:
            raise ValueError(
                "PDS labels should only have UTC times, but "
                f"this time has a timezone: {value}"
            )


class ISISEncoder(PVLEncoder):
    """An encoder for writing PVL text that can be parsed by the
    ISIS PVL text parser.

    The ISIS3 implementation (as of 3.9) of PVL/ODL (like) does not
    strictly follow any of the published standards. It was based
    on PDS3 ODL from the 1990s, but has several extensions adopted
    from existing and prior data sets from ISIS2, PDS, JAXA, ISRO,
    ..., and extensions used only within ISIS files (cub, net). This
    is one of the reasons using ISIS cube files or PVL text written by
    ISIS as an archive format has been strongly discouraged.

    Since there is no specification, only a detailed analysis of
    the ISIS software that parses and writes its PVL text would
    yield a strategy for parsing it.  This encoder is most likely the
    least reliable for that reason.  We welcome bug reports to help
    extend our coverage of this flavor of PVL text.

    :param grammar: defaults to pvl.grammar.ISISGrammar().
    :param decoder: defaults to pvl.decoder.PVLDecoder().
    :param end_delimiter: defaults to False.
    :param newline: defaults to '\\\\n'.
    """

    def __init__(
        self,
        grammar=None,
        decoder=None,
        indent=2,
        width=80,
        aggregation_end=True,
        end_delimiter=False,
        newline="\n",
    ):

        if grammar is None:
            grammar = ISISGrammar()

        if decoder is None:
            decoder = PVLDecoder(grammar)

        super().__init__(
            grammar,
            decoder,
            indent,
            width,
            aggregation_end,
            end_delimiter,
            newline,
        )



#!/usr/bin/env python
"""This module has new tests for the pvl decoder functions."""

# Copyright 2019, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import unittest

from pvl.parser import PVLParser, ParseError, OmniParser, EmptyValueAtLine
from pvl.lexer import lexer as Lexer
from pvl.lexer import LexerError
from pvl.collections import Quantity, PVLModule, PVLGroup, PVLObject


class TestParse(unittest.TestCase):
    def setUp(self):
        self.p = PVLParser()

    # def test_broken_assignment(self):
    #     self.assertRaises(DecodeError,
    #                       self.d.broken_assignment, 'foo', 0)

    #     self.d.strict = False
    #     empty = EmptyValueAtLine(1)
    #     self.assertEqual(empty, self.d.broken_assignment('foo', 0))

    # def test_parse_iterable(self):
    #     def pv(s, idx):
    #         (t, _, _) = s[idx:-1].partition(',')
    #         v = t.strip()
    #         i = s.find(v, idx)
    #         return v, i + len(v)

    #     with patch('pvl.decoder.PVLDecoder.parse_value', side_effect=pv):
    #         i = '( a, b, c, d )'
    #         v = ['a', 'b', 'c', 'd']
    #         self.assertEqual((v, len(i)),
    #                          self.d.parse_iterable(i, 0, '(', ')'))

    def test_parse_begin_aggregation_statement(self):
        pairs = (
            ("GROUP = name next", "GROUP", "name"),
            ("OBJECT=name next", "OBJECT", "name"),
            (
                "BEGIN_GROUP /*c1*/ = /*c2*/ name /*c3*/ next",
                "BEGIN_GROUP",
                "name",
            ),
        )
        for x in pairs:
            with self.subTest(pair=x):
                tokens = Lexer(x[0])
                # next_t = x[1].split()[-1]
                self.assertEqual(
                    (x[1], x[2]),
                    self.p.parse_begin_aggregation_statement(tokens),
                )

        tokens = Lexer("Not-a-Begin-Aggregation-Statement = name")
        self.assertRaises(
            ValueError, self.p.parse_begin_aggregation_statement, tokens
        )

        strings = ("GROUP equals name", "GROUP = 5")
        for s in strings:
            with self.subTest(string=s):
                tokens = Lexer(s)
                self.assertRaises(
                    ValueError,
                    self.p.parse_begin_aggregation_statement,
                    tokens,
                )

    def test_parse_end_aggregation(self):
        groups = (
            ("END_GROUP", "GROUP", "name"),
            ("END_GROUP = name", "BEGIN_GROUP", "name"),
            ("END_OBJECT /*c1*/ =   \n name", "OBJECT", "name"),
        )
        for g in groups:
            with self.subTest(groups=g):
                tokens = Lexer(g[0])
                self.assertIsNone(
                    self.p.parse_end_aggregation(g[1], g[2], tokens)
                )

        bad_groups = (
            ("END_GROUP", "OBJECT", "name"),
            ("END_GROUP = foo", "BEGIN_GROUP", "name"),
        )
        for g in bad_groups:
            with self.subTest(groups=g):
                tokens = Lexer(g[0])
                self.assertRaises(
                    ValueError,
                    self.p.parse_end_aggregation,
                    g[1],
                    g[2],
                    tokens,
                )

        tokens = Lexer("END_GROUP = ")
        self.assertRaises(
            StopIteration,
            self.p.parse_end_aggregation,
            "BEGIN_GROUP",
            "name",
            tokens,
        )

    def test_parse_around_equals(self):
        strings = ("=", " = ", "/*c1*/ = /*c2*/")
        for s in strings:
            with self.subTest(string=s):
                tokens = Lexer(s)
                self.assertIsNone(self.p.parse_around_equals(tokens))
        bad_strings = ("f", " f ")
        for s in bad_strings:
            with self.subTest(string=s):
                tokens = Lexer(s)
                self.assertRaises(
                    ValueError, self.p.parse_around_equals, tokens
                )
        tokens = Lexer("")
        self.assertRaises(ParseError, self.p.parse_around_equals, tokens)

        tokens = Lexer(" = f")
        self.p.parse_around_equals(tokens)
        f = next(tokens)
        self.assertEqual(f, "f")

    def test_parse_units(self):
        pairs = (
            (Quantity(5, "m"), "<m>"),
            (Quantity(5, "m"), "< m >"),
            (Quantity(5, "m /* comment */"), "< m /* comment */>"),
            (Quantity(5, "m\nfoo"), "< m\nfoo >"),
        )
        for p in pairs:
            with self.subTest(pairs=p):
                tokens = Lexer(p[1])
                self.assertEqual(p[0], self.p.parse_units(5, tokens))

    def test_parse_set(self):
        pairs = (
            ({"a", "b", "c"}, "{a,b,c}"),
            ({"a", "b", "c"}, "{ a, b, c }"),
            ({"a", "b", "c"}, "{ a, /* random */b, c }"),
            ({"a", frozenset(["x", "y"]), "c"}, "{ a, {x,y}, c }"),
        )
        for p in pairs:
            with self.subTest(pairs=p):
                tokens = Lexer(p[1])
                self.assertEqual(p[0], self.p.parse_set(tokens))

    def test_parse_sequence(self):
        pairs = (
            (["a", "b", "c"], "(a,b,c)"),
            (["a", "b", "c"], "( a, b, c )"),
            (["a", "b", "c"], "( a, /* random */b, c )"),
            (["a", ["x", "y"], "c"], "( a, (x,y), c )"),
        )
        for p in pairs:
            with self.subTest(pairs=p):
                tokens = Lexer(p[1])
                self.assertEqual(p[0], self.p.parse_sequence(tokens))

    def test_parse_WSC_until(self):
        triplets = (("   stop <units>", "stop", True),)
        for t in triplets:
            with self.subTest(triplet=t):
                tokens = Lexer(t[0])
                self.assertEqual(t[2], self.p.parse_WSC_until(t[1], tokens))

    def test_parse_value(self):
        pairs = (
            ("(a,b,c)", ["a", "b", "c"]),
            ("{ a, b, c }", {"a", "b", "c"}),
            ("2001-01-01", datetime.date(2001, 1, 1)),
            ("2#0101#", 5),
            ("-79", -79),
            ("Unquoted", "Unquoted"),
            ('"Quoted"', "Quoted"),
            ("Null", None),
            ("TRUE", True),
            ("false", False),
            ("9 <planets>", Quantity(9, "planets")),
        )
        for p in pairs:
            with self.subTest(pairs=p):
                tokens = Lexer(p[0])
                self.assertEqual(p[1], self.p.parse_value(tokens))

    def test_parse_assignment_statement(self):
        pairs = (
            ("a=b", "a", "b"),
            ("a =\tb", "a", "b"),
            ("a /*comment*/ = +80", "a", 80),
            ("a = b c = d", "a", "b"),
            ("a = b; c = d", "a", "b"),
        )
        for p in pairs:
            with self.subTest(pairs=p):
                tokens = Lexer(p[0])
                self.assertEqual(
                    (p[1], p[2]), self.p.parse_assignment_statement(tokens)
                )

        tokens = Lexer("empty = 2##")
        self.assertRaises(
            LexerError, self.p.parse_assignment_statement, tokens
        )

    def test_parse_aggregation_block(self):
        groups = (
            (
                "GROUP = name bob = uncle END_GROUP",
                ("name", PVLGroup(bob="uncle")),
            ),
            (
                "GROUP = name OBJECT = uncle name = bob END_OBJECT END_GROUP",
                ("name", PVLGroup(uncle=PVLObject(name="bob"))),
            ),
            (
                "GROUP = name bob = uncle END_GROUP = name next = token",
                ("name", PVLGroup(bob="uncle")),
            ),
        )
        for g in groups:
            with self.subTest(groups=g):
                tokens = Lexer(g[0])
                self.assertEqual(g[1], self.p.parse_aggregation_block(tokens))

        bad_blocks = (
            "Group = name bob = uncle END_OBJECT",
            "GROUP= name = bob = uncle END_GROUP",
            "",
        )
        for b in bad_blocks:
            with self.subTest(block=b):
                tokens = Lexer(b)
                self.assertRaises(
                    ValueError, self.p.parse_aggregation_block, tokens
                )

    def test_parse_end_statement(self):
        strings = (
            "END;",
            "End ; ",
            "End /*comment*/",
            "END next",
            "END",
            "end",
            "END ",
            "END\n\n",
        )
        for g in strings:
            with self.subTest(groups=g):
                tokens = Lexer(g)
                # top = Lexer(g)
                # for t in top:
                #     print(f'token : "{t}"')
                self.assertIsNone(self.p.parse_end_statement(tokens))

        tokens = Lexer("the_end")
        self.assertRaises(ValueError, self.p.parse_end_statement, tokens)

    def test_parse_module(self):
        groups = (
            ("a = b c = d END", PVLModule(a="b", c="d")),
            (
                "a =b GROUP = g f=g END_GROUP END",
                PVLModule(a="b", g=PVLGroup(f="g")),
            ),
            ("GROUP = g f=g END_GROUP END", PVLModule(g=PVLGroup(f="g"))),
            (
                "GROUP = g f=g END_GROUP a = b OBJECT = o END_OBJECT END",
                PVLModule(g=PVLGroup(f="g"), a="b", o=PVLObject()),
            ),
        )
        for g in groups:
            with self.subTest(groups=g):
                tokens = Lexer(g[0])
                self.assertEqual(g[1], self.p.parse_module(tokens))

        tokens = Lexer("blob")
        self.assertRaises(ParseError, self.p.parse_module, tokens)

        tokens = Lexer("blob =")
        self.assertRaises(ParseError, self.p.parse_module, tokens)

        tokens = Lexer("GROUP GROUP")
        self.assertRaises(LexerError, self.p.parse_module, tokens)

        tokens = Lexer("BEGIN_OBJECT = foo END_OBJECT = bar")
        self.assertRaises(ValueError, self.p.parse_module, tokens)

        tokens = Lexer(
            """ mixed = 'mixed"\\'quotes'
                           number = '123' """
        )
        self.assertRaises(LexerError, self.p.parse_module, tokens)

        # tokens = Lexer('blob = foo = bar')
        # self.p.parse_module(tokens)

    def test_parse(self):
        groups = (
            ("a = b c = d END", PVLModule(a="b", c="d")),
            (
                "a =b GROUP = g f=g END_GROUP END",
                PVLModule(a="b", g=PVLGroup(f="g")),
            ),
            ("GROUP = g f=g END_GROUP END", PVLModule(g=PVLGroup(f="g"))),
            (
                "GROUP = g f=g END_GROUP a = b OBJECT = o END_OBJECT END",
                PVLModule(g=PVLGroup(f="g"), a="b", o=PVLObject()),
            ),
        )
        for g in groups:
            with self.subTest(groups=g):
                self.assertEqual(g[1], self.p.parse(g[0]))

        self.assertRaises(ParseError, self.p.parse, "blob")


class TestOmni(unittest.TestCase):
    def setUp(self):
        self.p = OmniParser()

    def test_parse_module_post_hook(self):
        m = PVLModule(a="b")
        tokens = Lexer("c = d", g=self.p.grammar, d=self.p.decoder)
        self.assertRaises(Exception, self.p.parse_module_post_hook, m, tokens)

        self.p.doc = "a = b = d"
        m = PVLModule(a="b")
        tokens = Lexer("= d", g=self.p.grammar, d=self.p.decoder)
        mod = PVLModule(a=EmptyValueAtLine(0), b="d")
        self.assertEqual(
            (mod, False), self.p.parse_module_post_hook(m, tokens)
        )

        self.p.doc = "a = b ="
        m = PVLModule(a="b")
        tokens = Lexer("=", g=self.p.grammar, d=self.p.decoder)
        mod = PVLModule(a=EmptyValueAtLine(0), b=EmptyValueAtLine(0))
        self.assertEqual(
            (mod, False), self.p.parse_module_post_hook(m, tokens)
        )

    def test_comments(self):
        some_pvl = """
        /* comment on line */
        # here is a line comment
        /* here is a multi-
        line comment */
        foo = bar /* comment at end of line */
        weird/* in the */=/*middle*/comments
        baz = bang # end line comment
        End
    """

        self.assertEqual(
            PVLModule(foo="bar", weird="comments", baz="bang"),
            self.p.parse(some_pvl),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='pvl',
    version='1.2.0',
    description=(
        'Python implementation for PVL (Parameter Value Language) '
        'parsing and encoding.'
    ),
    long_description=readme + '\n\n' + history,
    author='The PlanetaryPy Developers',
    maintainer="Ross Beyer",
    maintainer_email="rbeyer@rossbeyer.net",
    url='https://github.com/planetarypy/pvl',
    packages=[
        'pvl',
    ],
    package_dir={'pvl':
                 'pvl'},
    include_package_data=True,
    license="BSD",
    zip_safe=False,
    keywords='pvl',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Text Processing'
    ],
    extras_require={
        "allopts": ["astropy", "multidict", "pint", "python-dateutil"],
        "dateutil":  ["python-dateutil"],
        "multidict": ["multidict"],
        "quantities": ["astropy", "pint"],
    },
    entry_points={"console_scripts": [
        "pvl_translate = pvl.pvl_translate:main",
        "pvl_validate= pvl.pvl_validate:main",
    ], }
)



# -*- coding: utf-8 -*-
"""Parameter Value Language decoder.

The definition of PVL used in this module is based on the Consultive
Committee for Space Data Systems, and their Parameter Value
Language Specification (CCSD0006 and CCSD0008), CCSDS 6441.0-B-2,
referred to as the Blue Book with a date of June 2000.

A decoder deals with converting strings given to it (typically
by the parser) to the appropriate Python type.
"""
# Copyright 2015, 2017, 2019-2021, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import re
from datetime import datetime, timedelta, timezone
from decimal import InvalidOperation
from itertools import repeat, chain
from warnings import warn

from .grammar import PVLGrammar, ODLGrammar, PDSGrammar
from .collections import Quantity
from .exceptions import QuantityError


def for_try_except(exception, function, *iterable):
    """Return the result of the first successful application of *function*
    to an element of *iterable*.  If the *function* raises an Exception
    of type *exception*, it will continue to the next item of *iterable*.
    If there are no successful applications an Exception of type
    *exception* will be raised.

    If additional *iterable* arguments are passed, *function* must
    take that many arguments and is applied to the items from
    all iterables in parallel (like ``map()``). With multiple iterables,
    the iterator stops when the shortest iterable is exhausted.
    """
    for tup in zip(*iterable):
        try:
            return function(*tup)
        except exception:
            pass

    raise exception


class PVLDecoder(object):
    """A decoder based on the rules in the CCSDS-641.0-B-2 'Blue Book'
    which defines the PVL language.

    :param grammar: defaults to a :class:`pvl.grammar.PVLGrammar`, but can
        be any object that implements the :class:`pvl.grammar` interface.

    :param quantity_cls: defaults to :class:`pvl.collections.Quantity`, but
        could be any class object that takes two arguments, where the
        first is the value, and the second is the units value.

    :param real_cls: defaults to :class:`float`, but could be any class object
        that can be constructed from a `str` object.
    """

    def __init__(self, grammar=None, quantity_cls=None, real_cls=None):
        self.errors = []

        if grammar is None:
            self.grammar = PVLGrammar()
        elif isinstance(grammar, PVLGrammar):
            self.grammar = grammar
        else:
            raise Exception

        if quantity_cls is None:
            self.quantity_cls = Quantity
        else:
            self.quantity_cls = quantity_cls

        if real_cls is None:
            self.real_cls = float
        else:
            self.real_cls = real_cls

    def decode(self, value: str):
        """Returns a Python object based on *value*."""
        return self.decode_simple_value(value)

    def decode_simple_value(self, value: str):
        """Returns a Python object based on *value*, assuming
        that *value* can be decoded as a PVL Simple Value::

         <Simple-Value> ::= (<Date-Time> | <Numeric> | <String>)
        """
        if value.casefold() == self.grammar.none_keyword.casefold():
            return None

        if value.casefold() == self.grammar.true_keyword.casefold():
            return True

        if value.casefold() == self.grammar.false_keyword.casefold():
            return False

        for d in (
            self.decode_quoted_string,
            self.decode_non_decimal,
            self.decode_decimal,
            self.decode_datetime,
        ):
            try:
                return d(value)
            except ValueError:
                pass

        return self.decode_unquoted_string(value)

    def decode_unquoted_string(self, value: str) -> str:
        """Returns a Python ``str`` if *value* can be decoded
        as an unquoted string, based on this decoder's grammar.
        Raises a ValueError otherwise.
        """
        for coll in (
            ("a comment", chain.from_iterable(self.grammar.comments)),
            ("some whitespace", self.grammar.whitespace),
            ("a special character", self.grammar.reserved_characters),
        ):
            for item in coll[1]:
                if item in value:
                    raise ValueError(
                        "Expected a Simple Value, but encountered "
                        f'{coll[0]} in "{self}": "{item}".'
                    )

        agg_keywords = self.grammar.aggregation_keywords.items()
        for kw in chain.from_iterable(agg_keywords):
            if kw.casefold() == value.casefold():
                raise ValueError(
                    "Expected a Simple Value, but encountered "
                    f'an aggregation keyword: "{value}".'
                )

        for es in self.grammar.end_statements:
            if es.casefold() == value.casefold():
                raise ValueError(
                    "Expected a Simple Value, but encountered "
                    f'an End-Statement: "{value}".'
                )

        # This try block is going to look illogical.  But the decode
        # rules for Unquoted Strings spell out the things that they
        # cannot be, so if it *can* be a datetime, then it *can't* be
        # an Unquoted String, which is why we raise if it succeeds,
        # and pass if it fails:
        try:
            self.decode_datetime(value)
            raise ValueError
        except ValueError:
            pass

        return str(value)

    def decode_quoted_string(self, value: str) -> str:
        """Returns a Python ``str`` if *value* begins and ends
        with matching quote characters based on this decoder's
        grammar.  Raises ValueError otherwise.
        """
        for q in self.grammar.quotes:
            if value.startswith(q) and value.endswith(q) and len(value) > 1:
                return str(value[1:-1])
        raise ValueError(f'The object "{value}" is not a PVL Quoted String.')

    def decode_decimal(self, value: str):
        """Returns a Python ``int`` or ``self.real_cls`` object, as appropriate
        based on *value*.  Raises a ValueError otherwise.
        """
        # Returns int or real_cls
        try:
            return int(value, base=10)
        except ValueError:
            try:
                return self.real_cls(str(value))
            except InvalidOperation as err:
                raise ValueError from err

    def decode_non_decimal(self, value: str) -> int:
        """Returns a Python ``int`` as decoded from *value*
        on the assumption that *value* conforms to a
        non-decimal integer value as defined by this decoder's
        grammar, raises ValueError otherwise.
        """
        # Non-Decimal (Binary, Hex, and Octal)
        for nd_re in (
            self.grammar.binary_re,
            self.grammar.octal_re,
            self.grammar.hex_re,
        ):
            match = nd_re.fullmatch(value)
            if match is not None:
                d = match.groupdict("")
                return int(d["sign"] + d["non_decimal"], base=int(d["radix"]))
        raise ValueError

    def decode_datetime(self, value: str):  # noqa: C901
        """Takes a string and attempts to convert it to the appropriate
        Python ``datetime`` ``time``, ``date``, or ``datetime``
        type based on this decoder's grammar, or in one case, a ``str``.

        The PVL standard allows for the seconds value to range
        from zero to 60, so that the 60 can accommodate leap
        seconds.  However, the Python ``datetime`` classes don't
        support second values for more than 59 seconds.

        If a time with 60 seconds is encountered, it will not be
        returned as a datetime object (since that is not representable
        via Python datetime objects), but simply as a string.

        The user can then then try and use the ``time`` module
        to parse this string into a ``time.struct_time``.  We
        chose not to do this with pvl because ``time.struct_time``
        is a full *datetime* like object, even if it parsed
        only a *time* like object, the year, month, and day
        values in the ``time.struct_time`` would default, which
        could be misleading.

        Alternately, the pvl.grammar.PVLGrammar class contains
        two regexes: ``leap_second_Ymd_re`` and ``leap_second_Yj_re``
        which could be used along with the ``re.match`` object's
        ``groupdict()`` function to extract the string representations
        of the various numerical values, cast them to the appropriate
        numerical types, and do something useful with them.
        """
        try:
            # datetime.date objects will always be naive, so just return:
            return for_try_except(
                ValueError,
                datetime.strptime,
                repeat(value),
                self.grammar.date_formats,
            ).date()
        except ValueError:
            # datetime.time and datetime.datetime might be either:
            d = None
            try:
                d = for_try_except(
                    ValueError,
                    datetime.strptime,
                    repeat(value),
                    self.grammar.time_formats,
                ).time()
            except ValueError:
                try:
                    d = for_try_except(
                        ValueError,
                        datetime.strptime,
                        repeat(value),
                        self.grammar.datetime_formats,
                    )
                except ValueError:
                    pass
            if d is not None:
                if d.utcoffset() is None:
                    if value.endswith("Z"):
                        return d.replace(tzinfo=timezone.utc)
                    elif self.grammar.default_timezone is not None:
                        return d.replace(tzinfo=self.grammar.default_timezone)
                return d

        # if we can regex a 60-second time, return str
        if self.is_leap_seconds(value):
            return str(value)
        else:
            raise ValueError

    def is_leap_seconds(self, value: str) -> bool:
        """Returns True if *value* is a time that matches the
        grammar's definition of a leap seconds time (a time string with
        a value of 60 for the seconds value).  False otherwise."""
        for r in (
            self.grammar.leap_second_Ymd_re,
            self.grammar.leap_second_Yj_re,
        ):
            if r is not None and r.fullmatch(value) is not None:
                return True
        else:
            return False

    def decode_quantity(self, value, unit):
        """Returns a Python object that represents a value with
           an associated unit, based on the values provided via
           *value* and *unit*.  This function creates an object
           based on the decoder's *quantity_cls*.
        """
        try:
            return self.quantity_cls(value, str(unit))
        except ValueError as err:
            raise QuantityError(err)


class ODLDecoder(PVLDecoder):
    """A decoder based on the rules in the PDS3 Standards Reference
    (version 3.8, 27 Feb 2009) Chapter 12: Object Description
    Language Specification and Usage.

    Extends PVLDecoder, and if *grammar* is not specified, it will
    default to an ODLGrammar() object.
    """

    def __init__(self, grammar=None, quantity_cls=None, real_cls=None):
        self.errors = []

        if grammar is None:
            grammar = ODLGrammar()

        super().__init__(
            grammar=grammar,
            quantity_cls=quantity_cls,
            real_cls=real_cls
        )

    def decode_datetime(self, value: str):
        """Extends parent function to also deal with datetimes
        and times with a time zone offset.

        If it cannot, it will raise a ValueError.
        """

        try:
            return super().decode_datetime(value)
        except ValueError:
            # if there is a +HH:MM or a -HH:MM suffix that
            # can be stripped, then we're in business.
            # Otherwise ...
            match = re.fullmatch(
                r"(?P<dt>.+?)"  # the part before the sign
                r"(?P<sign>[+-])"  # required sign
                r"(?P<hour>0?[0-9]|1[0-2])"  # 0 to 12
                fr"(?:{self.grammar._M_frag})?",  # Minutes
                value,
            )
            if match is not None:
                gd = match.groupdict(default=0)
                dt = super().decode_datetime(gd["dt"])
                offset = timedelta(
                    hours=int(gd["hour"]), minutes=int(gd["minute"])
                )
                if gd["sign"] == "-":
                    offset = -1 * offset
                return dt.replace(tzinfo=timezone(offset))
            raise ValueError

    def decode_non_decimal(self, value: str) -> int:
        """Extends parent function by allowing the wider variety of
        radix values that ODL permits over PVL.
        """
        match = self.grammar.nondecimal_re.fullmatch(value)
        if match is not None:
            d = match.groupdict("")
            return int(d["sign"] + d["non_decimal"], base=int(d["radix"]))
        raise ValueError

    def decode_quoted_string(self, value: str) -> str:
        """Extends parent function because the
        ODL specification allows for a dash (-) line continuation
        character that results in the dash, the line end, and any
        leading whitespace on the next line to be removed.  It also
        allows for a sequence of format effectors surrounded by
        spacing characters to be collapsed to a single space.
        """
        s = super().decode_quoted_string(value)

        # Deal with dash (-) continuation:
        # sp = ''.join(self.grammar.spacing_characters)
        fe = "".join(self.grammar.format_effectors)
        ws = "".join(self.grammar.whitespace)
        nodash = re.sub(fr"-[{fe}][{ws}]*", "", s)

        # Originally thought that only format effectors surrounded
        # by whitespace was to be collapsed
        # foo = re.sub(fr'[{sp}]*[{fe}]+[{sp}]*', ' ', nodash)

        # But really it collapses all whitespace and strips lead and trail.
        return re.sub(fr"[{ws}]+", " ", nodash.strip(ws))

    def decode_unquoted_string(self, value: str) -> str:
        """Extends parent function to provide the extra enforcement that only
        ODL Identifier text may be unquoted as a value.
        """
        s = super().decode_unquoted_string(value)

        if self.is_identifier(s):
            return s
        else:
            raise ValueError(
                f"Only text that qualifies as an ODL Identifier may be "
                f"unquoted as a value, and '{s}' is not."
            )

    @staticmethod
    def is_identifier(value):
        """Returns true if *value* is an ODL Identifier, false otherwise.

        An ODL Identifier is composed of letters, digits, and underscores.
        The first character must be a letter, and the last must not
        be an underscore.
        """
        if isinstance(value, str):
            if len(value) == 0:
                return False

            try:
                # Ensure we're dealing with ASCII
                value.encode(encoding="ascii")

                # value can't start with a letter or end with an underbar
                if not value[0].isalpha() or value.endswith("_"):
                    return False

                for c in value:
                    if not (c.isalpha() or c.isdigit() or c == "_"):
                        return False
                else:
                    return True

            except UnicodeError:
                return False
        else:
            return False


class PDSLabelDecoder(ODLDecoder):
    """A decoder based on the rules in the PDS3 Standards Reference
    (version 3.8, 27 Feb 2009) Chapter 12: Object Description
    Language Specification and Usage.

    Extends ODLDecoder, and if *grammar* is not specified, it will
    default to a PDS3Grammar() object.
    """

    def __init__(self, grammar=None, quantity_cls=None):
        self.errors = []

        if grammar is None:
            super().__init__(grammar=PDSGrammar(), quantity_cls=quantity_cls)
        else:
            super().__init__(grammar=grammar, quantity_cls=quantity_cls)

    def decode_datetime(self, value: str):
        """Overrides parent function since PDS3 forbids a timezone
        specification, and times with a precision more than miliseconds.

        If it cannot decode properly, it will raise a ValueError.
        """

        t = super(ODLDecoder, self).decode_datetime(value)

        if (
            hasattr(t, "microsecond")
            and t.microsecond != round(t.microsecond / 1000) * 1000
        ):
            raise ValueError(
                f"The PDS specification does not allow time values with"
                f"precision greater than miliseconds, and this has "
                f"microsecond precision: {t}."
            )

        return t


class OmniDecoder(ODLDecoder):
    """A permissive decoder that attempts to parse all forms of
    "PVL" that are thrown at it.

    Extends ODLDecoder.
    """

    def decode_non_decimal(self, value: str) -> int:
        """Extends parent function by allowing a plus or
        minus sign to be in two different positions
        in a non-decimal number, since PVL has one
        specification, and ODL has another.
        """
        # Non-Decimal with a variety of radix values and sign
        # positions.
        match = self.grammar.nondecimal_re.fullmatch(value)
        if match is not None:
            d = match.groupdict("")
            if "second_sign" in d:
                if d["sign"] != "" and d["second_sign"] != "":
                    raise ValueError(
                        f'The non-decimal value, "{value}", ' "has two signs."
                    )
                elif d["sign"] != "":
                    sign = d["sign"]
                else:
                    sign = d["second_sign"]
            else:
                sign = d["sign"]

            return int(sign + d["non_decimal"], base=int(d["radix"]))
        raise ValueError

    def decode_datetime(self, value: str):
        """Returns an appropriate Python datetime time, date, or datetime
        object by using the 3rd party dateutil library (if present)
        to parse an ISO 8601 datetime string in *value*.  If it cannot,
        or the dateutil library is not present, it will raise a
        ValueError.
        """

        try:
            return super().decode_datetime(value)
        except ValueError:
            try:
                from dateutil.parser import isoparser

                isop = isoparser()

                if len(value) > 3 and value[-2] == "+" and value[-1].isdigit():
                    # This technically means that we accept slightly more
                    # formats than ISO 8601 strings, since under that
                    # specification, two digits after the '+' are required
                    # for an hour offset, but ODL doesn't have this
                    # requirement.  If we find only one digit, we'll
                    # just assume it means an hour and insert a zero so
                    # that it can be parsed.
                    tokens = value.rpartition("+")
                    value = tokens[0] + "+0" + tokens[-1]

                try:
                    return isop.parse_isodate(value)
                except ValueError:
                    try:
                        return isop.parse_isotime(value)
                    except ValueError:
                        return isop.isoparse(value)

            except ImportError:
                warn(
                    "The dateutil library is not present, so more "
                    "exotic date and time formats beyond the PVL/ODL "
                    "set cannot be parsed.",
                    ImportWarning,
                )

            raise ValueError

    def decode_unquoted_string(self, value: str) -> str:
        """Overrides parent function since the ODLDecoder has a more narrow
        definition of what is allowable as an unquoted string than the
        PVLDecoder does.
        """
        return super(ODLDecoder, self).decode_unquoted_string(value)



#!/usr/bin/env python
"""This module has new tests for the pvl decoder functions."""

# Copyright 2019, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pvl.grammar import PVLGrammar, OmniGrammar

import pvl.lexer as Lexer


class TestInternal(unittest.TestCase):
    def test_prev_char(self):
        self.assertIsNone(Lexer._prev_char("foo", 0))
        self.assertEqual("f", Lexer._prev_char("foo", 1))

    def test_next_char(self):
        self.assertIsNone(Lexer._next_char("foo", 2))
        self.assertEqual("b", Lexer._next_char("fob", 1))

    def test_prepare_comment_tuples(self):
        d = dict(
            single_comments=dict(),
            multi_comments=(("/*", "*/"),),
            multi_chars=set("/*"),
            chars=set("/*"),
        )
        self.assertEqual(d, Lexer._prepare_comment_tuples((("/*", "*/"),)))

        d = dict(
            single_comments={"#": "\n"},
            multi_comments=(("/*", "*/"),),
            multi_chars=set("/*"),
            chars=set("/*#"),
        )
        self.assertEqual(
            d, Lexer._prepare_comment_tuples((("/*", "*/"), ("#", "\n")))
        )


class TestLexComments(unittest.TestCase):
    def test_lex_multichar_comments(self):
        pfn = dict(state=Lexer.Preserve.FALSE, end=None)
        pcom = dict(state=Lexer.Preserve.COMMENT, end="*/")
        self.assertEqual(
            ("", pfn), Lexer.lex_multichar_comments("a", "b", "c", "", pfn)
        )
        self.assertEqual(
            ("", pfn), Lexer.lex_multichar_comments("/", "*", "c", "", pfn)
        )
        self.assertEqual(
            ("", pfn), Lexer.lex_multichar_comments("/", "b", "*", "", pfn)
        )
        self.assertEqual(
            ("/", pfn), Lexer.lex_multichar_comments("/", "b", "c", "", pfn)
        )
        self.assertEqual(
            ("*", pcom), Lexer.lex_multichar_comments("*", "b", "c", "", pcom)
        )
        self.assertEqual(
            ("/*", pcom), Lexer.lex_multichar_comments("*", "/", "c", "", pfn)
        )
        self.assertEqual(
            ("*/", pfn), Lexer.lex_multichar_comments("*", "c", "/", "", pcom)
        )

        self.assertRaises(
            ValueError,
            Lexer.lex_multichar_comments,
            "a",
            "b",
            "c",
            "",
            pfn,
            tuple(),
        )
        self.assertRaises(
            NotImplementedError,
            Lexer.lex_multichar_comments,
            "a",
            "b",
            "c",
            "",
            pfn,
            (("/*", "*/"), ("|*", "*|")),
        )

    def test_lex_singlechar_comments(self):
        pfn = dict(state=Lexer.Preserve.FALSE, end=None)
        phash = dict(state=Lexer.Preserve.COMMENT, end="\n")
        self.assertEqual(
            ("", pfn), Lexer.lex_singlechar_comments("a", "", pfn, {"k": "v"})
        )
        self.assertEqual(
            ("#", phash),
            Lexer.lex_singlechar_comments("#", "", pfn, {"#": "\n"}),
        )
        self.assertEqual(
            ("#\n", pfn),
            Lexer.lex_singlechar_comments("\n", "#", phash, {"#": "\n"}),
        )

    def test_lex_comment(self):
        pfn = dict(state=Lexer.Preserve.FALSE, end=None)
        pcom = dict(state=Lexer.Preserve.COMMENT, end="*/")
        self.assertEqual(
            ("", pfn),
            Lexer.lex_comment(
                "a",
                "b",
                "c",
                "",
                pfn,
                dict(
                    single_comments={"k": "v"},
                    multi_comments=(("/*", "*/"),),
                    multi_chars={"/", "*"},
                ),
            ),
        )

        self.assertEqual(
            ("/*", pcom),
            Lexer.lex_comment(
                "*",
                "/",
                "c",
                "",
                pfn,
                dict(
                    single_comments={"k": "v"},
                    multi_comments=(("/*", "*/"),),
                    multi_chars={"/", "*"},
                ),
            ),
        )


class TestLexer(unittest.TestCase):
    def setUp(self):
        def get_tokens(s):
            tokens = list()
            lex = Lexer.lexer(s)
            for t in lex:
                # print(f'yields: {t}')
                tokens.append(t)
            return tokens

        self.get_tokens = get_tokens

    def test_plain(self):
        s = "This is a test."
        tokens = s.split()
        out = self.get_tokens(s)
        self.assertEqual(tokens, out)

    def test_newline(self):
        s = "This \n is  a\ttest."
        tokens = ["This", "is", "a", "test."]
        out = self.get_tokens(s)
        self.assertEqual(tokens, out)

    def test_reserved(self):
        s = "Te=st"
        tokens = ["Te", "=", "st"]
        out = self.get_tokens(s)
        self.assertEqual(tokens, out)

    def test_comment(self):
        s = "There is a /* comment */"
        tokens = ["There", "is", "a", "/* comment */"]
        out = self.get_tokens(s)
        self.assertEqual(tokens, out)

        s = "/* At */ the beginning"
        tokens = ["/* At */", "the", "beginning"]
        out = self.get_tokens(s)
        self.assertEqual(tokens, out)

        s = "In/*the*/middle."
        tokens = ["In", "/*the*/", "middle."]
        out = self.get_tokens(s)
        self.assertEqual(tokens, out)

    def test_custom_comment(self):
        def get_tokens(s):
            tokens = list()
            g = PVLGrammar()
            g.comments = (("/*", "*/"), ("#", "\n"))
            lex = Lexer.lexer(s, g=g)
            for t in lex:
                # print(f'yields: {t}')
                tokens.append(t)
            return tokens

        s = "There is a # comment"
        tokens = ["There", "is", "a", "# comment"]
        out = get_tokens(s)
        self.assertEqual(tokens, out)

        s = "There is a # comment \n then more"
        tokens = ["There", "is", "a", "# comment \n", "then", "more"]
        out = get_tokens(s)
        self.assertEqual(tokens, out)

        s = "# Leading \n then \n more"
        tokens = ["# Leading \n", "then", "more"]
        out = get_tokens(s)
        self.assertEqual(tokens, out)

    def test_omni_comment(self):
        def get_tokens(s):
            tokens = list()
            lex = Lexer.lexer(s, g=OmniGrammar())
            for t in lex:
                # print(f'yields: {t}')
                tokens.append(t)
            return tokens

        s = """
        /* comment on line */
        # here is a line comment
        /* here is a multi-
        line comment */
        foo = bar /* comment at end of line */
        weird/* in the */=/*middle*/comments
        baz = bang # end line comment
        End
    """
        out = get_tokens(s)
        self.assertEqual(
            [
                "/* comment on line */",
                "# here is a line comment\n",
                "/* here is a multi-\n        line comment */",
                "foo",
                "=",
                "bar",
                "/* comment at end of line */",
                "weird",
                "/* in the */",
                "=",
                "/*middle*/",
                "comments",
                "baz",
                "=",
                "bang",
                "# end line comment\n",
                "End",
            ],
            out,
        )

    def test_quotes(self):
        pairs = (
            (
                '"This is quoted." Notquoted',
                ['"This is quoted."', "Notquoted"],
            ),
            (
                "\"These 'are' inner\" quotes",
                ["\"These 'are' inner\"", "quotes"],
            ),
            (
                "'Other # reserved <chars>' outside",
                ["'Other # reserved <chars>'", "outside"],
            ),
        )
        for p in pairs:
            with self.subTest(pairs=p):
                out = self.get_tokens(p[0])
                self.assertEqual(p[1], out)

    # def test_foo(self):
    #     out = self.get_tokens("""single = 'single"quotes'
    #     mixed = 'mixed"\\'quotes'
    #     number = '123'
    # """)

    def test_numeric(self):
        pairs = (
            ("Number: +79", ["Number:", "+79"]),
            ("Binary: +2#0101#", ["Binary:", "+2#0101#"]),
            ("Scientific_notation: 2e+2", ["Scientific_notation:", "2e+2"]),
        )
        # ('Binary: +2#0101#', ['Binary:', '+2', '#', '0101', '#']))
        for p in pairs:
            with self.subTest(pairs=p):
                out = self.get_tokens(p[0])
                self.assertEqual(p[1], out)

    def test_send(self):
        s = "One Two Three"
        tokens = Lexer.lexer(s)
        for t in tokens:
            if t == "Two":
                tokens.send(t)  # return the Token to the generator
                break

        self.assertEqual("Two", next(tokens))

    def test_lex_char(self):
        g = PVLGrammar()
        p = dict(state=Lexer.Preserve.FALSE, end="end")
        self.assertEqual(
            ("a", p),
            Lexer.lex_char(
                "a",
                "b",
                "c",
                "",
                p,
                g,
                dict(
                    chars={"k", "v", "/", "*"},
                    single_comments={"k": "v"},
                    multi_chars={"/", "*"},
                ),
            ),
        )

    def test_lexer_recurse(self):
        def foo(tokens):
            two = list()
            for t in tokens:
                if t == "f":
                    break
                two.append(t)
            return two

        lex = Lexer.lexer("a b c d e f g h")
        one = list()
        for t in lex:
            if t == "c":
                two = foo(lex)
            else:
                one.append(t)

        self.assertEqual(["a", "b", "g", "h"], one)
        self.assertEqual(["d", "e"], two)
